"""Tests for load-aware scheduler and load metrics (Task 2.7)."""

from __future__ import annotations

from sagellm_kv_cache.load_aware_scheduler import LoadAwareScheduler, PlacementStrategy
from sagellm_kv_cache.load_metrics import DeviceLoadMetrics, LoadMetricsCollector
from sagellm_kv_cache.models import RequestType, SchedulerRequest


class TestLoadMetrics:
    """Test suite for load metrics."""

    def test_device_load_metrics_basic(self) -> None:
        metrics = DeviceLoadMetrics.create(
            "cuda:0", kv_tokens_capacity=1000, memory_capacity_mb=100
        )
        metrics.update_kv_usage(300)
        metrics.update_request_count(active=2, queued=1)
        metrics.update_memory_usage(20)

        assert metrics.kv_utilization() == 0.3
        assert metrics.memory_utilization() == 0.2
        assert metrics.total_requests() == 3
        assert metrics.available_kv_tokens() == 700

    def test_load_metrics_collector(self) -> None:
        collector = LoadMetricsCollector()
        collector.register_device("cuda:0", kv_tokens_capacity=1000)
        collector.register_device("cuda:1", kv_tokens_capacity=1000)
        collector.update_kv_usage("cuda:0", 500)

        assert set(collector.list_devices()) == {"cuda:0", "cuda:1"}
        assert collector.get_metrics("cuda:0").kv_tokens_used == 500


class TestLoadAwareScheduler:
    """Test suite for load-aware placement strategies."""

    def _collector(self) -> LoadMetricsCollector:
        collector = LoadMetricsCollector()
        collector.register_device("cuda:0", kv_tokens_capacity=1000)
        collector.register_device("cuda:1", kv_tokens_capacity=1000)
        return collector

    def test_lowest_kv_strategy(self) -> None:
        collector = self._collector()
        collector.update_kv_usage("cuda:0", 700)
        collector.update_kv_usage("cuda:1", 200)

        scheduler = LoadAwareScheduler(collector, strategy=PlacementStrategy.LOWEST_KV)
        selected = scheduler.select_device(["cuda:0", "cuda:1"], kv_tokens_required=100)

        assert selected == "cuda:1"

    def test_shortest_queue_strategy(self) -> None:
        collector = self._collector()
        collector.update_request_count("cuda:0", active=4, queued=2)
        collector.update_request_count("cuda:1", active=1, queued=1)

        scheduler = LoadAwareScheduler(collector, strategy=PlacementStrategy.SHORTEST_QUEUE)
        selected = scheduler.select_device(["cuda:0", "cuda:1"], kv_tokens_required=100)

        assert selected == "cuda:1"

    def test_disable_load_awareness_fallback_round_robin(self) -> None:
        collector = self._collector()
        scheduler = LoadAwareScheduler(
            collector,
            strategy=PlacementStrategy.LOWEST_KV,
            enable_load_awareness=False,
        )

        first = scheduler.select_device(["cuda:0", "cuda:1"])
        second = scheduler.select_device(["cuda:0", "cuda:1"])

        assert (first, second) == ("cuda:0", "cuda:1")

    def test_request_device_affinity_is_honored(self) -> None:
        collector = self._collector()
        scheduler = LoadAwareScheduler(collector, strategy=PlacementStrategy.LOWEST_KV)

        request = SchedulerRequest(
            request_id="req-affinity",
            trace_id="trace-affinity",
            request_type=RequestType.PREFILL,
            prompt_len=16,
            max_new_tokens=8,
            kv_tokens_required=24,
            device_affinity="cuda:1",
        )

        selected = scheduler.select_device_for_request(request)
        assert selected == "cuda:1"



"""测试分布式池化接口的预留功能

验证：
1. KVHandle 支持 node_rank 字段
2. KVPool 支持 node_rank 参数
3. 分布式池化接口可导入（虽未实现）
"""

from __future__ import annotations

import pytest

from sagellm_kv_cache import (
    GlobalKVPoolManager,
    KVHandle,
    NodeContribution,
    RemoteKVAccessor,
)
from sagellm_kv_cache.pool import KVPool


class TestNodeRankSupport:
    """测试 node_rank 字段支持"""

    def test_kv_handle_with_node_rank(self) -> None:
        """测试 KVHandle 支持 node_rank 参数"""
        handle = KVHandle.create(num_tokens=128, device="cpu", node_rank=2)

        assert handle.node_rank == 2
        assert handle.num_tokens == 128

    def test_kv_handle_without_node_rank(self) -> None:
        """测试 KVHandle 默认 node_rank=None（本地）"""
        handle = KVHandle.create(num_tokens=128, device="cpu")

        assert handle.node_rank is None

    def test_kv_handle_is_local(self) -> None:
        """测试 is_local() 方法"""
        handle_local = KVHandle.create(num_tokens=128, device="cpu", node_rank=None)
        handle_node_0 = KVHandle.create(num_tokens=128, device="cpu", node_rank=0)
        handle_node_1 = KVHandle.create(num_tokens=128, device="cpu", node_rank=1)

        # node_rank=None 总是本地
        assert handle_local.is_local(current_rank=0) is True
        assert handle_local.is_local(current_rank=1) is True

        # node_rank=0 只在 rank 0 本地
        assert handle_node_0.is_local(current_rank=0) is True
        assert handle_node_0.is_local(current_rank=1) is False

        # node_rank=1 只在 rank 1 本地
        assert handle_node_1.is_local(current_rank=0) is False
        assert handle_node_1.is_local(current_rank=1) is True

    def test_kv_pool_with_node_rank(self) -> None:
        """测试 KVPool 支持 node_rank 参数"""
        pool = KVPool(max_tokens=1024, node_rank=2)

        assert pool.node_rank == 2
        assert pool.max_tokens == 1024

    def test_kv_pool_without_node_rank(self) -> None:
        """测试 KVPool 默认 node_rank=None（单机模式）"""
        pool = KVPool(max_tokens=1024)

        assert pool.node_rank is None

    def test_kv_pool_alloc_sets_node_rank(self) -> None:
        """测试 KVPool.alloc() 自动设置 node_rank"""
        pool = KVPool(max_tokens=1024, node_rank=3)
        handle = pool.alloc(128, dtype="fp16", layout="contiguous", device="cpu")

        # 应该自动继承 pool 的 node_rank
        assert handle.node_rank == 3

    def test_kv_pool_alloc_without_node_rank(self) -> None:
        """测试 KVPool.alloc() 在单机模式下不设置 node_rank"""
        pool = KVPool(max_tokens=1024, node_rank=None)
        handle = pool.alloc(128, dtype="fp16", layout="contiguous", device="cpu")

        # 应该保持 None
        assert handle.node_rank is None


class TestDistributedInterfacesImportable:
    """测试分布式接口可导入（虽未实现）"""

    def test_global_kv_pool_manager_importable(self) -> None:
        """测试 GlobalKVPoolManager 可导入"""
        assert GlobalKVPoolManager is not None
        assert hasattr(GlobalKVPoolManager, "__abstractmethods__")

    def test_node_contribution_importable(self) -> None:
        """测试 NodeContribution 可导入"""
        assert NodeContribution is not None

    def test_remote_kv_accessor_importable(self) -> None:
        """测试 RemoteKVAccessor 可导入"""
        assert RemoteKVAccessor is not None
        assert hasattr(RemoteKVAccessor, "__abstractmethods__")

    def test_node_contribution_validation(self) -> None:
        """测试 NodeContribution 验证逻辑"""
        # 正常配置
        contrib = NodeContribution(
            node_rank=0,
            total_memory_gb=100.0,
            contributed_memory_gb=40.0,
            reserved_memory_gb=60.0,
            storage_tiers=["gpu", "cpu"],
        )
        assert contrib.node_rank == 0
        assert contrib.contributed_memory_gb == 40.0

        # 超额分配应失败
        with pytest.raises(ValueError, match="Contributed.*Reserved.*Total"):
            NodeContribution(
                node_rank=0,
                total_memory_gb=100.0,
                contributed_memory_gb=60.0,
                reserved_memory_gb=50.0,  # 60 + 50 > 100
                storage_tiers=["gpu"],
            )

        # 负数应失败
        with pytest.raises(ValueError, match="must be non-negative"):
            NodeContribution(
                node_rank=0,
                total_memory_gb=100.0,
                contributed_memory_gb=-10.0,
                reserved_memory_gb=60.0,
                storage_tiers=["gpu"],
            )


class TestDistributedInterfaceDocumentation:
    """测试分布式接口的文档完整性"""

    def test_global_kv_pool_manager_has_methods(self) -> None:
        """测试 GlobalKVPoolManager 定义了所需方法"""
        required_methods = [
            "__init__",
            "alloc",
            "free",
            "get",
            "get_global_stats",
            "rebalance",
        ]

        for method in required_methods:
            assert hasattr(GlobalKVPoolManager, method), f"Missing method: {method}"

    def test_remote_kv_accessor_has_methods(self) -> None:
        """测试 RemoteKVAccessor 定义了所需方法"""
        required_methods = ["fetch", "prefetch", "is_remote"]

        for method in required_methods:
            assert hasattr(RemoteKVAccessor, method), f"Missing method: {method}"



"""Tests for Prefix Cache (Task 2.1)"""

import pytest

from sagellm_kv_cache.prefix_cache import PrefixCache


class TestPrefixCache:
    """Test suite for PrefixCache."""

    def test_initialization(self):
        """Test basic initialization."""
        cache = PrefixCache(block_size=16)
        assert cache.block_size == 16
        assert len(cache) == 0

        with pytest.raises(ValueError):
            PrefixCache(block_size=0)

    def test_compute_block_hash_deterministic(self):
        """Test that block hash is deterministic."""
        cache = PrefixCache(block_size=16)
        tokens = list(range(16))

        hash1 = cache.compute_block_hash(tokens)
        hash2 = cache.compute_block_hash(tokens)

        assert hash1 == hash2
        assert isinstance(hash1, bytes)
        assert len(hash1) == 32  # SHA256 = 32 bytes

    def test_compute_block_hash_different_tokens(self):
        """Test that different tokens produce different hashes."""
        cache = PrefixCache(block_size=16)
        tokens1 = list(range(16))
        tokens2 = list(range(1, 17))

        hash1 = cache.compute_block_hash(tokens1)
        hash2 = cache.compute_block_hash(tokens2)

        assert hash1 != hash2

    def test_compute_block_hash_wrong_size(self):
        """Test error when token count doesn't match block_size."""
        cache = PrefixCache(block_size=16)
        tokens = list(range(10))  # Wrong size

        with pytest.raises(ValueError, match="doesn't match block_size"):
            cache.compute_block_hash(tokens)

    def test_compute_block_hashes(self):
        """Test computing hashes for multiple blocks."""
        cache = PrefixCache(block_size=16)
        tokens = list(range(50))  # 3 complete blocks + 2 tokens

        hashes = cache.compute_block_hashes(tokens)

        assert len(hashes) == 3  # Only complete blocks
        assert all(isinstance(h, bytes) for h in hashes)
        assert all(len(h) == 32 for h in hashes)

    def test_insert_and_lookup_hit(self):
        """Test successful cache lookup."""
        cache = PrefixCache(block_size=16)
        tokens = list(range(48))
        hashes = cache.compute_block_hashes(tokens)
        blocks = [{"block_id": i, "tokens": 16} for i in range(3)]

        # Insert
        cache.insert(hashes, blocks)
        assert len(cache) == 3

        # Lookup - full match
        hit_blocks, num_tokens = cache.lookup(hashes)
        assert len(hit_blocks) == 3
        assert num_tokens == 48

        # Check stats
        stats = cache.get_stats()
        assert stats["hit_count"] == 1
        assert stats["miss_count"] == 0
        assert stats["hit_rate"] == 1.0

    def test_lookup_partial_match(self):
        """Test partial prefix match."""
        cache = PrefixCache(block_size=16)
        tokens = list(range(48))
        hashes = cache.compute_block_hashes(tokens)
        blocks = [{"block_id": i} for i in range(3)]

        # Insert only first 2 blocks
        cache.insert(hashes[:2], blocks[:2])

        # Lookup with 3 blocks - should match first 2
        hit_blocks, num_tokens = cache.lookup(hashes)
        assert len(hit_blocks) == 2
        assert num_tokens == 32

    def test_lookup_miss(self):
        """Test cache miss."""
        cache = PrefixCache(block_size=16)
        hashes = [b"hash1", b"hash2"]

        # Lookup without insert
        hit_blocks, num_tokens = cache.lookup(hashes)
        assert len(hit_blocks) == 0
        assert num_tokens == 0

        stats = cache.get_stats()
        assert stats["miss_count"] == 1

    def test_invalidate(self):
        """Test cache invalidation."""
        from uuid import uuid4

        cache = PrefixCache(block_size=16)
        handle_id = uuid4()
        hashes = [b"hash1", b"hash2"]
        blocks = [
            {"block_id": 0, "handle_id": handle_id},
            {"block_id": 1, "handle_id": handle_id},
        ]

        cache.insert(hashes, blocks)
        assert len(cache) == 2

        # Invalidate
        removed = cache.invalidate(handle_id)
        assert removed == 2
        assert len(cache) == 0

    def test_invalidate_partial(self):
        """Test invalidating some but not all blocks."""
        from uuid import uuid4

        cache = PrefixCache(block_size=16)
        handle_id1 = uuid4()
        handle_id2 = uuid4()

        hashes = [b"hash1", b"hash2", b"hash3"]
        blocks = [
            {"block_id": 0, "handle_id": handle_id1},
            {"block_id": 1, "handle_id": handle_id2},
            {"block_id": 2, "handle_id": handle_id1},
        ]

        cache.insert(hashes, blocks)

        # Invalidate handle_id1
        removed = cache.invalidate(handle_id1)
        assert removed == 2

        # hash2 should still be cached
        hit_blocks, _ = cache.lookup([hashes[1]])
        assert len(hit_blocks) == 1

    def test_clear(self):
        """Test clearing entire cache."""
        cache = PrefixCache(block_size=16)
        hashes = [b"hash1", b"hash2"]
        blocks = [{"block_id": 0}, {"block_id": 1}]

        cache.insert(hashes, blocks)
        assert len(cache) == 2

        cache.clear()
        assert len(cache) == 0

    def test_insert_duplicate_hash(self):
        """Test inserting multiple blocks with same hash."""
        cache = PrefixCache(block_size=16)
        hash1 = b"hash1"

        # Insert twice with same hash
        cache.insert([hash1], [{"block_id": 0}])
        cache.insert([hash1], [{"block_id": 1}])

        # Should have 1 entry but 2 blocks
        stats = cache.get_stats()
        assert stats["num_entries"] == 1
        assert stats["num_blocks"] == 2

    def test_stats_tracking(self):
        """Test statistics tracking."""
        cache = PrefixCache(block_size=16)
        hashes = [b"hash1", b"hash2"]
        blocks = [{"block_id": 0}, {"block_id": 1}]

        cache.insert(hashes, blocks)

        # Hit
        cache.lookup(hashes)
        # Miss
        cache.lookup([b"hash3"])

        stats = cache.get_stats()
        assert stats["hit_count"] == 1
        assert stats["miss_count"] == 1
        assert stats["hit_rate"] == 0.5
        assert stats["insert_count"] == 2

    def test_repr(self):
        """Test string representation."""
        cache = PrefixCache(block_size=16)
        repr_str = repr(cache)

        assert "PrefixCache" in repr_str
        assert "block_size=16" in repr_str
        assert "hit_rate" in repr_str


class TestPrefixCacheLRU:
    """Test suite for Prefix Cache LRU eviction (issue #55)."""

    def test_lru_initialization(self):
        """Test LRU initialization."""
        cache = PrefixCache(block_size=16, max_cached_blocks=10, enable_lru=True)
        assert cache.block_size == 16
        assert cache.max_cached_blocks == 10
        assert cache.enable_lru is True

        stats = cache.get_stats()
        assert stats["enable_lru"] is True
        assert stats["max_cached_blocks"] == 10

    def test_lru_disabled(self):
        """Test cache with LRU disabled."""
        cache = PrefixCache(block_size=16, enable_lru=False)
        assert cache.enable_lru is False

        # evict_by_count should fail when LRU is disabled
        with pytest.raises(RuntimeError, match="LRU eviction not enabled"):
            cache.evict_by_count(1)

    def test_lru_access_updates_order(self):
        """Test that lookup updates LRU order."""
        import time

        cache = PrefixCache(block_size=16, max_cached_blocks=3, enable_lru=True)
        hashes = [b"hash1", b"hash2", b"hash3"]
        blocks = [{"block_id": i} for i in range(3)]

        # Insert all blocks
        for h, b in zip(hashes, blocks):
            cache.insert([h], [b])
            time.sleep(0.01)  # Ensure different timestamps

        # Access hash1 to make it most recent
        cache.lookup([hashes[0]])

        # Now hash2 should be least recently used
        # Insert a new block - should evict hash2
        cache.insert([b"hash4"], [{"block_id": 4}])

        # hash1 and hash3 should still exist
        hit_blocks1, _ = cache.lookup([hashes[0]])
        assert len(hit_blocks1) == 1

        hit_blocks3, _ = cache.lookup([hashes[2]])
        assert len(hit_blocks3) == 1

        # hash2 should be evicted
        hit_blocks2, _ = cache.lookup([hashes[1]])
        assert len(hit_blocks2) == 0

    def test_lru_eviction_on_capacity(self):
        """Test automatic eviction when capacity is reached."""
        cache = PrefixCache(block_size=16, max_cached_blocks=2, enable_lru=True)
        hashes = [b"hash1", b"hash2", b"hash3"]
        blocks = [{"block_id": i} for i in range(3)]

        # Insert 2 blocks - should be fine
        cache.insert(hashes[:2], blocks[:2])
        assert len(cache) == 2

        # Insert 3rd block - should trigger eviction of hash1 (oldest)
        cache.insert([hashes[2]], [blocks[2]])

        # hash1 should be evicted (oldest)
        hit_blocks1, _ = cache.lookup([hashes[0]])
        assert len(hit_blocks1) == 0

        # hash2 and hash3 should still exist
        hit_blocks2, _ = cache.lookup([hashes[1]])
        assert len(hit_blocks2) == 1

        hit_blocks3, _ = cache.lookup([hashes[2]])
        assert len(hit_blocks3) == 1

        # Check eviction count
        stats = cache.get_stats()
        assert stats["evict_count"] >= 1

    def test_manual_eviction(self):
        """Test manual eviction via evict_by_count."""
        cache = PrefixCache(block_size=16, max_cached_blocks=10, enable_lru=True)
        hashes = [b"hash1", b"hash2", b"hash3"]
        blocks = [{"block_id": i} for i in range(3)]

        cache.insert(hashes, blocks)
        assert len(cache) == 3

        # Manually evict 2 blocks
        evicted = cache.evict_by_count(2)
        assert evicted >= 2

        # Should have at most 1 block left
        stats = cache.get_stats()
        assert stats["num_blocks"] <= 1

    def test_lru_stats(self):
        """Test that LRU stats are properly tracked."""
        cache = PrefixCache(block_size=16, max_cached_blocks=2, enable_lru=True)
        hashes = [b"hash1", b"hash2", b"hash3"]
        blocks = [{"block_id": i} for i in range(3)]

        # Insert 3 blocks - should trigger eviction
        for h, b in zip(hashes, blocks):
            cache.insert([h], [b])

        stats = cache.get_stats()
        assert stats["insert_count"] == 3
        assert stats["evict_count"] >= 1
        assert stats["enable_lru"] is True

    def test_lru_clear(self):
        """Test that clear removes LRU tracker entries."""
        cache = PrefixCache(block_size=16, max_cached_blocks=10, enable_lru=True)
        hashes = [b"hash1", b"hash2"]
        blocks = [{"block_id": i} for i in range(2)]

        cache.insert(hashes, blocks)
        assert len(cache) == 2
        assert len(cache._lru_tracker) == 2

        cache.clear()
        assert len(cache) == 0
        assert len(cache._lru_tracker) == 0

    def test_lru_invalidate(self):
        """Test that invalidate also updates LRU tracker."""
        from uuid import uuid4

        cache = PrefixCache(block_size=16, max_cached_blocks=10, enable_lru=True)
        handle_id = uuid4()
        hashes = [b"hash1", b"hash2"]
        blocks = [
            {"block_id": 0, "handle_id": handle_id},
            {"block_id": 1, "handle_id": handle_id},
        ]

        cache.insert(hashes, blocks)
        assert len(cache) == 2
        assert len(cache._lru_tracker) == 2

        # Invalidate
        cache.invalidate(handle_id)
        assert len(cache) == 0
        assert len(cache._lru_tracker) == 0



"""Tests for Eviction Policy (Task 2.3)"""

import time

import pytest

from sagellm_kv_cache.errors import KVAllPinnedError, KVNoVictimError
from sagellm_kv_cache.eviction import (
    EvictionManager,
    FIFOEviction,
    LRUEviction,
    PriorityEviction,
    TTLEviction,
)
from sagellm_kv_cache.models import EvictionPolicy, KVHandle


class TestLRUEviction:
    """Test suite for LRU eviction strategy."""

    def test_lru_basic(self):
        """Test basic LRU eviction."""
        policy = LRUEviction()

        # Create handles with different access times
        h1 = KVHandle.create(num_tokens=100)
        h2 = KVHandle.create(num_tokens=100)
        h3 = KVHandle.create(num_tokens=100)

        h1.last_access = 1.0
        h2.last_access = 2.0
        h3.last_access = 3.0

        # Select victim - should pick h1 (oldest)
        victims = policy.select_victim([h1, h2, h3], 100)

        assert len(victims) == 1
        assert victims[0] == h1

    def test_lru_multiple_victims(self):
        """Test LRU with multiple victims needed."""
        policy = LRUEviction()

        h1 = KVHandle.create(num_tokens=100)
        h2 = KVHandle.create(num_tokens=100)
        h3 = KVHandle.create(num_tokens=100)

        h1.last_access = 1.0
        h2.last_access = 2.0
        h3.last_access = 3.0

        # Need 250 tokens - should evict h1 and h2
        victims = policy.select_victim([h1, h2, h3], 250)

        assert len(victims) == 3  # Need all 3 to get 300 tokens
        assert victims[0] == h1
        assert victims[1] == h2

    def test_lru_insufficient_tokens(self):
        """Test LRU when not enough tokens available."""
        policy = LRUEviction()

        h1 = KVHandle.create(num_tokens=50)
        h2 = KVHandle.create(num_tokens=50)

        # Need 200 tokens but only 100 available
        with pytest.raises(KVNoVictimError):
            policy.select_victim([h1, h2], 200)

    def test_lru_no_candidates(self):
        """Test LRU with no candidates."""
        policy = LRUEviction()

        with pytest.raises(KVNoVictimError):
            policy.select_victim([], 100)


class TestFIFOEviction:
    """Test suite for FIFO eviction strategy."""

    def test_fifo_basic(self):
        """Test basic FIFO eviction."""
        policy = FIFOEviction()

        h1 = KVHandle.create(num_tokens=100)
        h2 = KVHandle.create(num_tokens=100)
        h3 = KVHandle.create(num_tokens=100)

        h1.created_at = 1.0
        h2.created_at = 2.0
        h3.created_at = 3.0

        # Select victim - should pick h1 (oldest creation)
        victims = policy.select_victim([h1, h2, h3], 100)

        assert len(victims) == 1
        assert victims[0] == h1

    def test_fifo_vs_lru(self):
        """Test that FIFO uses creation time, not access time."""
        policy = FIFOEviction()

        h1 = KVHandle.create(num_tokens=100)
        h2 = KVHandle.create(num_tokens=100)

        # h1 created first but accessed later
        h1.created_at = 1.0
        h1.last_access = 10.0

        h2.created_at = 2.0
        h2.last_access = 5.0

        # FIFO should still pick h1 (older creation)
        victims = policy.select_victim([h1, h2], 100)
        assert victims[0] == h1


class TestEvictionManager:
    """Test suite for EvictionManager."""

    def test_initialization(self):
        """Test manager initialization."""
        manager = EvictionManager(policy=EvictionPolicy.LRU)
        assert manager.policy == EvictionPolicy.LRU

        manager = EvictionManager(policy=EvictionPolicy.FIFO)
        assert manager.policy == EvictionPolicy.FIFO

    def test_set_policy(self):
        """Test changing policy."""
        manager = EvictionManager(policy=EvictionPolicy.LRU)
        manager.set_policy(EvictionPolicy.FIFO)

        assert manager.policy == EvictionPolicy.FIFO

    def test_filter_pinned_handles(self):
        """Test that pinned handles are filtered out."""
        manager = EvictionManager(policy=EvictionPolicy.LRU)

        h1 = KVHandle.create(num_tokens=100)
        h2 = KVHandle.create(num_tokens=100)
        h3 = KVHandle.create(num_tokens=100)

        # Pin h1
        h1.pin()
        assert not h1.is_evictable()

        h1.last_access = 1.0
        h2.last_access = 2.0
        h3.last_access = 3.0

        # Should not select h1 even though it's oldest
        victims = manager.select_victims([h1, h2, h3], 100)

        assert h1 not in victims
        assert victims[0] == h2  # Next oldest

    def test_all_pinned_error(self):
        """Test error when all handles are pinned."""
        manager = EvictionManager()

        h1 = KVHandle.create(num_tokens=100)
        h2 = KVHandle.create(num_tokens=100)

        h1.pin()
        h2.pin()

        with pytest.raises(KVAllPinnedError):
            manager.select_victims([h1, h2], 100)

    def test_multiple_pin_counts(self):
        """Test handles with multiple pin counts."""
        manager = EvictionManager()

        h1 = KVHandle.create(num_tokens=100)
        h2 = KVHandle.create(num_tokens=100)

        # Pin h1 twice
        h1.pin()
        h1.pin()
        assert h1.pin_count == 2

        h1.last_access = 1.0
        h2.last_access = 2.0

        victims = manager.select_victims([h1, h2], 100)
        assert victims[0] == h2

        # Unpin once - still pinned
        h1.unpin()
        assert h1.pin_count == 1
        victims = manager.select_victims([h1, h2], 100)
        assert victims[0] == h2

        # Unpin again - now evictable
        h1.unpin()
        assert h1.pin_count == 0
        victims = manager.select_victims([h1, h2], 100)
        assert victims[0] == h1

    def test_repr(self):
        """Test string representation."""
        manager = EvictionManager(policy=EvictionPolicy.LRU)
        repr_str = repr(manager)

        assert "EvictionManager" in repr_str
        assert "lru" in repr_str  # Enum value is lowercase


class TestTTLEviction:
    """Test suite for TTL eviction strategy."""

    def test_ttl_basic_expired(self):
        """Test TTL eviction with expired handles."""
        policy = TTLEviction(ttl_seconds=60)

        h1 = KVHandle.create(num_tokens=100)
        h2 = KVHandle.create(num_tokens=100)

        # h1 expired (2 minutes ago), h2 not expired (30 seconds ago)
        h1.created_at = time.time() - 120
        h2.created_at = time.time() - 30

        # Should evict h1 first (expired)
        victims = policy.select_victim([h1, h2], 100)

        assert len(victims) == 1
        assert victims[0] == h1

    def test_ttl_all_expired(self):
        """Test TTL when all handles are expired."""
        policy = TTLEviction(ttl_seconds=60)

        h1 = KVHandle.create(num_tokens=100)
        h2 = KVHandle.create(num_tokens=100)
        h3 = KVHandle.create(num_tokens=100)

        # All expired at different times
        h1.created_at = time.time() - 200
        h2.created_at = time.time() - 150
        h3.created_at = time.time() - 100

        # Should evict oldest expired first
        victims = policy.select_victim([h1, h2, h3], 100)

        assert len(victims) == 1
        assert victims[0] == h1  # Oldest creation time

    def test_ttl_none_expired_fallback_to_lru(self):
        """Test TTL fallback to LRU when no handles are expired."""
        policy = TTLEviction(ttl_seconds=300)

        h1 = KVHandle.create(num_tokens=100)
        h2 = KVHandle.create(num_tokens=100)

        # Neither expired (within 5 minutes)
        current = time.time()
        h1.created_at = current - 60
        h2.created_at = current - 30

        # Set different access times
        h1.last_access = current - 10
        h2.last_access = current - 5

        # Should fall back to LRU (evict h1 with older access)
        victims = policy.select_victim([h1, h2], 100)

        assert len(victims) == 1
        assert victims[0] == h1

    def test_ttl_mixed_expired_and_fresh(self):
        """Test TTL with mix of expired and fresh handles."""
        policy = TTLEviction(ttl_seconds=60)

        h1 = KVHandle.create(num_tokens=50)
        h2 = KVHandle.create(num_tokens=50)
        h3 = KVHandle.create(num_tokens=50)
        h4 = KVHandle.create(num_tokens=50)

        current = time.time()

        # h1, h2 expired
        h1.created_at = current - 120
        h2.created_at = current - 100

        # h3, h4 not expired
        h3.created_at = current - 40
        h4.created_at = current - 20

        # Set access times
        h3.last_access = current - 30
        h4.last_access = current - 10

        # Need 150 tokens - should evict h1, h2 (expired) first, then h3 (oldest access)
        victims = policy.select_victim([h1, h2, h3, h4], 150)

        assert len(victims) == 3
        assert h1 in victims
        assert h2 in victims
        assert h3 in victims
        assert h4 not in victims

    def test_ttl_no_candidates(self):
        """Test TTL with no candidates."""
        policy = TTLEviction(ttl_seconds=60)

        with pytest.raises(KVNoVictimError):
            policy.select_victim([], 100)

    def test_ttl_policy_name(self):
        """Test TTL policy returns correct enum."""
        policy = TTLEviction()
        assert policy.get_policy_name() == EvictionPolicy.TTL


class TestPriorityEviction:
    """Test suite for Priority eviction strategy."""

    def test_priority_basic(self):
        """Test basic priority eviction."""
        policy = PriorityEviction()

        h1 = KVHandle.create(num_tokens=100)
        h2 = KVHandle.create(num_tokens=100)

        h1.metadata = {"priority": 1}
        h2.metadata = {"priority": 5}

        # Should evict h1 (lower priority)
        victims = policy.select_victim([h1, h2], 100)

        assert len(victims) == 1
        assert victims[0] == h1

    def test_priority_multiple_levels(self):
        """Test priority with multiple priority levels."""
        policy = PriorityEviction()

        h1 = KVHandle.create(num_tokens=100)
        h2 = KVHandle.create(num_tokens=100)
        h3 = KVHandle.create(num_tokens=100)

        h1.metadata = {"priority": 1}
        h2.metadata = {"priority": 3}
        h3.metadata = {"priority": 5}

        # Need 150 tokens - should evict h1 and h2 (200 tokens total)
        victims = policy.select_victim([h1, h2, h3], 150)

        assert len(victims) >= 2
        assert h1 in victims
        assert h2 in victims
        # h3 might be included to reach 150 tokens, but h1 and h2 should be first
        # Check that victims are ordered by priority
        assert victims[0] == h1  # Lowest priority first
        assert victims[1] == h2  # Next lowest priority

    def test_priority_default_priority(self):
        """Test handles without explicit priority get default."""
        policy = PriorityEviction(default_priority=5)

        h1 = KVHandle.create(num_tokens=100)
        h2 = KVHandle.create(num_tokens=100)

        h1.metadata = {"priority": 3}
        # h2 has no priority - should use default (5)

        # Should evict h1 (priority 3 < default 5)
        victims = policy.select_victim([h1, h2], 100)

        assert len(victims) == 1
        assert victims[0] == h1

    def test_priority_same_priority_lru_tiebreaker(self):
        """Test LRU tiebreaker for same priority."""
        policy = PriorityEviction()

        h1 = KVHandle.create(num_tokens=100)
        h2 = KVHandle.create(num_tokens=100)

        # Same priority
        h1.metadata = {"priority": 3}
        h2.metadata = {"priority": 3}

        # Different access times
        h1.last_access = 1.0
        h2.last_access = 2.0

        # Should evict h1 (older access)
        victims = policy.select_victim([h1, h2], 100)

        assert len(victims) == 1
        assert victims[0] == h1

    def test_priority_no_candidates(self):
        """Test priority with no candidates."""
        policy = PriorityEviction()

        with pytest.raises(KVNoVictimError):
            policy.select_victim([], 100)

    def test_priority_policy_name(self):
        """Test Priority policy returns correct enum."""
        policy = PriorityEviction()
        assert policy.get_policy_name() == EvictionPolicy.PRIORITY


class TestEvictionManagerExtended:
    """Extended tests for EvictionManager with new policies."""

    def test_manager_ttl_policy(self):
        """Test manager with TTL policy."""
        manager = EvictionManager(policy=EvictionPolicy.TTL, ttl_seconds=60)

        h1 = KVHandle.create(num_tokens=100)
        h2 = KVHandle.create(num_tokens=100)

        h1.created_at = time.time() - 120  # Expired
        h2.created_at = time.time() - 30  # Not expired

        victims = manager.select_victims([h1, h2], 100)

        assert h1 in victims
        assert h2 not in victims

    def test_manager_priority_policy(self):
        """Test manager with Priority policy."""
        manager = EvictionManager(policy=EvictionPolicy.PRIORITY, default_priority=0)

        h1 = KVHandle.create(num_tokens=100)
        h2 = KVHandle.create(num_tokens=100)

        h1.metadata = {"priority": 1}
        h2.metadata = {"priority": 10}

        victims = manager.select_victims([h1, h2], 100)

        assert h1 in victims
        assert h2 not in victims

    def test_policy_hot_swap(self):
        """Test hot-swapping eviction policies."""
        manager = EvictionManager(policy=EvictionPolicy.LRU)

        # Initially LRU
        assert manager.policy == EvictionPolicy.LRU

        # Swap to FIFO
        manager.set_policy(EvictionPolicy.FIFO)
        assert manager.policy == EvictionPolicy.FIFO

        # Swap to TTL
        manager.set_policy(EvictionPolicy.TTL)
        assert manager.policy == EvictionPolicy.TTL

        # Swap to Priority
        manager.set_policy(EvictionPolicy.PRIORITY)
        assert manager.policy == EvictionPolicy.PRIORITY


class TestEvictionPerformance:
    """Performance tests for eviction policies (验收标准: evict_ms < 5ms)."""

    def test_lru_performance_100_handles(self):
        """Test LRU performance with 100 handles (Year1: < 5ms)."""
        policy = LRUEviction()

        # Create 100 handles
        handles = []
        current = time.time()
        for i in range(100):
            h = KVHandle.create(num_tokens=100)
            h.last_access = current - i
            handles.append(h)

        # Measure eviction time
        start = time.perf_counter()
        victims = policy.select_victim(handles, 1000)  # Need 10 handles
        elapsed_ms = (time.perf_counter() - start) * 1000.0

        assert len(victims) == 10
        assert elapsed_ms < 5.0, f"LRU eviction took {elapsed_ms:.2f}ms (> 5ms threshold)"

    def test_ttl_performance_100_handles(self):
        """Test TTL performance with 100 handles (Year1: < 5ms)."""
        policy = TTLEviction(ttl_seconds=60)

        # Create 100 handles (50 expired, 50 not)
        handles = []
        current = time.time()
        for i in range(50):
            h = KVHandle.create(num_tokens=100)
            h.created_at = current - 120  # Expired
            h.last_access = current - i
            handles.append(h)
        for i in range(50):
            h = KVHandle.create(num_tokens=100)
            h.created_at = current - 30  # Not expired
            h.last_access = current - i
            handles.append(h)

        # Measure eviction time
        start = time.perf_counter()
        victims = policy.select_victim(handles, 1000)  # Need 10 handles
        elapsed_ms = (time.perf_counter() - start) * 1000.0

        assert len(victims) == 10
        assert elapsed_ms < 5.0, f"TTL eviction took {elapsed_ms:.2f}ms (> 5ms threshold)"

    def test_priority_performance_100_handles(self):
        """Test Priority performance with 100 handles (Year1: < 5ms)."""
        policy = PriorityEviction()

        # Create 100 handles with varying priorities
        handles = []
        current = time.time()
        for i in range(100):
            h = KVHandle.create(num_tokens=100)
            h.metadata = {"priority": i % 10}  # Priorities 0-9
            h.last_access = current - i
            handles.append(h)

        # Measure eviction time
        start = time.perf_counter()
        victims = policy.select_victim(handles, 1000)  # Need 10 handles
        elapsed_ms = (time.perf_counter() - start) * 1000.0

        assert len(victims) == 10
        assert elapsed_ms < 5.0, f"Priority eviction took {elapsed_ms:.2f}ms (> 5ms threshold)"

    def test_fifo_performance_100_handles(self):
        """Test FIFO performance with 100 handles (Year1: < 5ms)."""
        policy = FIFOEviction()

        # Create 100 handles
        handles = []
        current = time.time()
        for i in range(100):
            h = KVHandle.create(num_tokens=100)
            h.created_at = current - i
            handles.append(h)

        # Measure eviction time
        start = time.perf_counter()
        victims = policy.select_victim(handles, 1000)  # Need 10 handles
        elapsed_ms = (time.perf_counter() - start) * 1000.0

        assert len(victims) == 10
        assert elapsed_ms < 5.0, f"FIFO eviction took {elapsed_ms:.2f}ms (> 5ms threshold)"

    def test_manager_performance_with_pinned(self):
        """Test EvictionManager performance with pinned handles."""
        manager = EvictionManager(policy=EvictionPolicy.LRU)

        # Create 100 handles, pin 20 of them
        handles = []
        current = time.time()
        for i in range(100):
            h = KVHandle.create(num_tokens=100)
            h.last_access = current - i
            if i < 20:
                h.pin()
            handles.append(h)

        # Measure eviction time
        start = time.perf_counter()
        victims = manager.select_victims(handles, 1000)  # Need 10 handles
        elapsed_ms = (time.perf_counter() - start) * 1000.0

        assert len(victims) == 10
        # None of the pinned handles should be selected
        for v in victims:
            assert not v.is_pinned
        assert elapsed_ms < 5.0, f"Manager eviction took {elapsed_ms:.2f}ms (> 5ms threshold)"



"""Tests for sagellm_kv_cache.errors

测试 KV Cache 错误类的功能：
- 错误基类行为
- 各类错误的初始化与消息格式
- 可重试判断
- 序列化

Author: sageLLM Team (Agent 2)
"""

from __future__ import annotations

from uuid import uuid4

import pytest
from sagellm_protocol.kv import KVErrorCode

from sagellm_kv_cache.errors import (
    KVAllPinnedError,
    # 预算/资源错误
    KVBudgetExceededError,
    # 基类
    KVCacheError,
    # 缓存错误
    KVCacheMissError,
    KVChecksumMismatchError,
    KVConfigInvalidError,
    # 配置错误
    KVConfigMissingError,
    KVDoubleFreeError,
    KVEvictionFailedError,
    KVHandleNotFoundError,
    KVHandlePinnedError,
    # 句柄操作错误
    KVInvalidHandleError,
    KVMemoryExhaustedError,
    KVMetadataMismatchError,
    KVNotImplementedError,
    # 驱逐错误
    KVNoVictimError,
    KVPoolFullError,
    KVPrefixTooLongError,
    KVRefCountUnderflowError,
    # 传输错误
    KVTransferFailedError,
    KVTransferTimeoutError,
)


class TestKVCacheErrorBase:
    """测试 KVCacheError 基类"""

    def test_basic_initialization(self):
        """测试基类初始化"""
        error = KVCacheError(
            code=KVErrorCode.BUDGET_EXCEEDED,
            message="Test error message",
        )

        assert error.code == KVErrorCode.BUDGET_EXCEEDED
        assert error.message == "Test error message"
        assert error.context == {}
        assert error.cause is None

    def test_with_context(self):
        """测试带上下文的初始化"""
        ctx = {"key": "value", "count": 42}
        error = KVCacheError(
            code=KVErrorCode.INVALID_HANDLE,
            message="Error with context",
            context=ctx,
        )

        assert error.context == ctx

    def test_with_cause(self):
        """测试带原因的初始化"""
        original = ValueError("Original error")
        error = KVCacheError(
            code=KVErrorCode.TRANSFER_FAILED,
            message="Wrapped error",
            cause=original,
        )

        assert error.cause is original

    def test_to_dict(self):
        """测试 to_dict() 方法"""
        error = KVCacheError(
            code=KVErrorCode.BUDGET_EXCEEDED,
            message="Test message",
            context={"requested": 100},
        )
        data = error.to_dict()

        assert data["error_type"] == "KVCacheError"
        assert data["code"] == "KV_BUDGET_EXCEEDED"
        assert data["message"] == "Test message"
        assert data["context"] == {"requested": 100}
        assert "retryable" in data

    def test_is_retryable(self):
        """测试可重试判断"""
        # 不可重试
        error = KVCacheError(code=KVErrorCode.BUDGET_EXCEEDED, message="")
        assert not error.is_retryable()

        # 可重试（TRANSFER_TIMEOUT 是可重试的）
        error = KVCacheError(code=KVErrorCode.TRANSFER_TIMEOUT, message="")
        assert error.is_retryable()

    def test_str_repr(self):
        """测试字符串表示"""
        error = KVCacheError(
            code=KVErrorCode.INVALID_HANDLE,
            message="Handle not found",
        )

        assert "[KV_INVALID_HANDLE]" in str(error)
        assert "Handle not found" in str(error)
        assert "KVCacheError" in repr(error)


class TestBudgetErrors:
    """测试预算/资源错误"""

    def test_budget_exceeded_error(self):
        """测试 KVBudgetExceededError"""
        error = KVBudgetExceededError(requested=1000, available=500)

        assert error.code == KVErrorCode.BUDGET_EXCEEDED
        assert error.requested == 1000
        assert error.available == 500
        assert "1000" in error.message
        assert "500" in error.message
        assert "tokens" in error.message

    def test_budget_exceeded_error_with_bytes(self):
        """测试 KVBudgetExceededError 使用字节单位"""
        error = KVBudgetExceededError(requested=2048, available=1024, unit="bytes")

        assert error.unit == "bytes"
        assert "bytes" in error.message

    def test_memory_exhausted_error(self):
        """测试 KVMemoryExhaustedError"""
        error = KVMemoryExhaustedError(required_bytes=1024 * 1024, device="cuda:0")

        assert error.code == KVErrorCode.MEMORY_EXHAUSTED
        assert error.required_bytes == 1024 * 1024
        assert error.device == "cuda:0"
        assert "cuda:0" in error.message

    def test_pool_full_error(self):
        """测试 KVPoolFullError"""
        error = KVPoolFullError(total_blocks=256)

        assert error.code == KVErrorCode.POOL_FULL
        assert error.total_blocks == 256
        assert "256" in error.message


class TestHandleErrors:
    """测试句柄操作错误"""

    def test_invalid_handle_error(self):
        """测试 KVInvalidHandleError"""
        handle_id = uuid4()
        error = KVInvalidHandleError(handle_id=handle_id, reason="expired")

        assert error.code == KVErrorCode.INVALID_HANDLE
        assert str(handle_id) in error.handle_id
        assert "expired" in error.message

    def test_double_free_error(self):
        """测试 KVDoubleFreeError"""
        handle_id = "test-handle-123"
        error = KVDoubleFreeError(handle_id=handle_id)

        assert error.code == KVErrorCode.DOUBLE_FREE
        assert error.handle_id == handle_id
        assert "already been freed" in error.message

    def test_handle_pinned_error(self):
        """测试 KVHandlePinnedError"""
        error = KVHandlePinnedError(
            handle_id="h-123",
            operation="evict",
            pin_count=3,
        )

        assert error.code == KVErrorCode.HANDLE_PINNED
        assert error.operation == "evict"
        assert error.pin_count == 3
        assert "evict" in error.message
        assert "3" in error.message

    def test_handle_not_found_error(self):
        """测试 KVHandleNotFoundError"""
        error = KVHandleNotFoundError(handle_id="missing-handle")

        assert error.code == KVErrorCode.HANDLE_NOT_FOUND
        assert "missing-handle" in error.message

    def test_ref_count_underflow_error(self):
        """测试 KVRefCountUnderflowError"""
        error = KVRefCountUnderflowError(handle_id="h-456", current_count=0)

        assert error.code == KVErrorCode.REF_COUNT_UNDERFLOW
        assert error.current_count == 0
        assert "underflow" in error.message.lower()


class TestCacheErrors:
    """测试缓存错误"""

    def test_cache_miss_error(self):
        """测试 KVCacheMissError"""
        error = KVCacheMissError(prefix_hash="abc123", prefix_len=128)

        assert error.code == KVErrorCode.CACHE_MISS
        assert error.prefix_hash == "abc123"
        assert error.prefix_len == 128
        assert "abc123" in error.message

    def test_prefix_too_long_error(self):
        """测试 KVPrefixTooLongError"""
        error = KVPrefixTooLongError(prefix_len=2048, max_len=1024)

        assert error.code == KVErrorCode.PREFIX_TOO_LONG
        assert error.prefix_len == 2048
        assert error.max_len == 1024
        assert "2048" in error.message
        assert "1024" in error.message


class TestEvictionErrors:
    """测试驱逐错误"""

    def test_no_victim_error(self):
        """测试 KVNoVictimError"""
        error = KVNoVictimError(
            bytes_needed=10240,
            pinned_count=50,
            total_count=50,
        )

        assert error.code == KVErrorCode.NO_VICTIM
        assert error.bytes_needed == 10240
        assert error.pinned_count == 50
        assert "10240" in error.message
        assert "50/50" in error.message

    def test_eviction_failed_error(self):
        """测试 KVEvictionFailedError"""
        error = KVEvictionFailedError(handle_id="h-789", reason="in use by another process")

        assert error.code == KVErrorCode.EVICTION_FAILED
        assert "h-789" in error.handle_id
        assert "in use" in error.message

    def test_all_pinned_error(self):
        """测试 KVAllPinnedError"""
        error = KVAllPinnedError(candidate_count=32)

        assert error.code == KVErrorCode.ALL_PINNED
        assert error.candidate_count == 32
        assert "32" in error.message


class TestTransferErrors:
    """测试传输错误"""

    def test_transfer_failed_error(self):
        """测试 KVTransferFailedError"""
        transfer_id = uuid4()
        error = KVTransferFailedError(
            transfer_id=transfer_id,
            source="cuda:0",
            target="cuda:1",
            reason="NCCL connection failed",
        )

        assert error.code == KVErrorCode.TRANSFER_FAILED
        assert str(transfer_id) in error.transfer_id
        assert error.source == "cuda:0"
        assert error.target == "cuda:1"
        assert "NCCL" in error.message

    def test_transfer_timeout_error(self):
        """测试 KVTransferTimeoutError"""
        error = KVTransferTimeoutError(
            transfer_id="t-123",
            timeout_seconds=30.0,
            elapsed_seconds=45.5,
        )

        assert error.code == KVErrorCode.TRANSFER_TIMEOUT
        assert error.timeout_seconds == 30.0
        assert error.elapsed_seconds == 45.5
        assert error.is_retryable()  # 超时应该是可重试的
        assert "45.50" in error.message
        assert "30.00" in error.message

    def test_checksum_mismatch_error(self):
        """测试 KVChecksumMismatchError"""
        error = KVChecksumMismatchError(
            expected="abc123",
            actual="def456",
            transfer_id="t-789",
        )

        assert error.code == KVErrorCode.CHECKSUM_MISMATCH
        assert error.expected == "abc123"
        assert error.actual == "def456"
        assert "abc123" in error.message
        assert "def456" in error.message

    def test_metadata_mismatch_error(self):
        """测试 KVMetadataMismatchError"""
        error = KVMetadataMismatchError(
            field="num_tokens",
            expected=128,
            actual=256,
        )

        assert error.code == KVErrorCode.METADATA_MISMATCH
        assert error.field == "num_tokens"
        assert error.expected == 128
        assert error.actual == 256
        assert "num_tokens" in error.message


class TestConfigErrors:
    """测试配置错误"""

    def test_config_missing_error(self):
        """测试 KVConfigMissingError"""
        error = KVConfigMissingError(config_key="max_tokens")

        assert error.code == KVErrorCode.CONFIG_MISSING
        assert error.config_key == "max_tokens"
        assert "max_tokens" in error.message

    def test_config_invalid_error(self):
        """测试 KVConfigInvalidError"""
        error = KVConfigInvalidError(
            config_key="batch_size",
            value=-1,
            reason="must be positive",
        )

        assert error.code == KVErrorCode.CONFIG_INVALID
        assert error.config_key == "batch_size"
        assert error.value == -1
        assert "-1" in error.message
        assert "positive" in error.message

    def test_not_implemented_error(self):
        """测试 KVNotImplementedError"""
        error = KVNotImplementedError(feature="async transfer")

        assert error.code == KVErrorCode.NOT_IMPLEMENTED
        assert error.feature == "async transfer"
        assert "async transfer" in error.message
        assert "Fail-Fast" in error.message


class TestErrorCatching:
    """测试错误捕获与处理"""

    def test_catch_by_base_class(self):
        """测试通过基类捕获"""
        with pytest.raises(KVCacheError) as exc_info:
            raise KVBudgetExceededError(requested=100, available=50)

        error = exc_info.value
        assert error.code == KVErrorCode.BUDGET_EXCEEDED

    def test_catch_by_specific_class(self):
        """测试通过具体类捕获"""
        with pytest.raises(KVBudgetExceededError) as exc_info:
            raise KVBudgetExceededError(requested=200, available=100)

        error = exc_info.value
        assert error.requested == 200

    def test_multiple_error_types(self):
        """测试多种错误类型处理"""
        errors = [
            KVBudgetExceededError(requested=100, available=50),
            KVInvalidHandleError(handle_id="h-1"),
            KVNoVictimError(bytes_needed=1024),
            KVTransferFailedError(transfer_id="t-1", source="a", target="b"),
            KVConfigMissingError(config_key="key"),
        ]

        for error in errors:
            assert isinstance(error, KVCacheError)
            assert isinstance(error.code, KVErrorCode)
            assert len(error.message) > 0
            data = error.to_dict()
            assert "code" in data
            assert "message" in data


class TestErrorChaining:
    """测试错误链"""

    def test_error_with_cause(self):
        """测试带原因的错误"""
        original = ValueError("Original error")
        error = KVTransferFailedError(
            transfer_id="t-chain",
            source="src",
            target="dst",
            reason=str(original),
            context={"cause": str(original)},
        )

        assert "Original error" in error.message

    def test_nested_error_handling(self):
        """测试嵌套错误处理"""
        try:
            try:
                raise ValueError("Low level error")
            except ValueError as e:
                raise KVTransferFailedError(
                    transfer_id="t-nested",
                    source="a",
                    target="b",
                    reason=str(e),
                ) from e
        except KVTransferFailedError as error:
            assert error.__cause__ is not None
            assert "Low level error" in str(error.__cause__)



"""KV Cache Profiling Module

This module provides tools for profiling and analyzing KV cache access patterns.
"""

from __future__ import annotations

from .access_stats import AccessStatsCollector

__all__ = ["AccessStatsCollector"]



"""Tests for Lifetime Predictor (Task 2.5).

Tests cover:
- Feature extraction
- Historical predictor (main implementation)
- Prediction accuracy and performance
"""

from __future__ import annotations

import time

import pytest

from sagellm_kv_cache.models import KVHandle
from sagellm_kv_cache.predictor import (
    FeatureExtractor,
    HistoricalPredictor,
    PredictorFeatures,
)


class TestPredictorFeatures:
    """Test feature extraction."""

    def test_from_handle_basic(self) -> None:
        """Test basic feature extraction from KVHandle."""
        handle = KVHandle.create(num_tokens=128, device="cuda:0")
        handle.access_count = 5
        handle.last_access = time.time()

        features = PredictorFeatures.from_handle(handle, hit_len=64)

        assert features.handle_id == handle.handle_id
        assert features.num_tokens == 128
        assert features.access_count == 5
        assert features.hit_len == 64
        assert features.device == "cuda:0"
        assert features.is_pinned is False

    def test_from_handle_with_session(self) -> None:
        """Test feature extraction with session ID."""
        handle = KVHandle.create(
            num_tokens=256,
            session_id="session-123",
            request_id="req-456",
        )

        features = PredictorFeatures.from_handle(handle)

        assert features.session_id == "session-123"
        assert features.request_id == "req-456"

    def test_to_dict_and_from_dict(self) -> None:
        """Test serialization and deserialization."""
        handle = KVHandle.create(num_tokens=64)
        features = PredictorFeatures.from_handle(handle)

        data = features.to_dict()
        restored = PredictorFeatures.from_dict(data)

        assert restored.handle_id == features.handle_id
        assert restored.num_tokens == features.num_tokens


class TestFeatureExtractor:
    """Test batch feature extraction."""

    def test_extract_batch(self) -> None:
        """Test extracting features from multiple handles."""
        handles = [KVHandle.create(num_tokens=i * 100) for i in range(1, 4)]
        extractor = FeatureExtractor()

        features = extractor.extract_batch(handles)

        assert len(features) == 3
        assert features[0].num_tokens == 100
        assert features[1].num_tokens == 200
        assert features[2].num_tokens == 300

    def test_extract_batch_with_hit_lens(self) -> None:
        """Test extraction with prefix hit lengths."""
        handles = [KVHandle.create(num_tokens=200) for _ in range(3)]
        hit_lens = [10, 20, 30]
        extractor = FeatureExtractor()

        features = extractor.extract_batch(handles, hit_lens=hit_lens)

        assert features[0].hit_len == 10
        assert features[1].hit_len == 20
        assert features[2].hit_len == 30

    def test_extract_batch_hit_lens_mismatch(self) -> None:
        """Test error when hit_lens length doesn't match handles."""
        handles = [KVHandle.create(num_tokens=100) for _ in range(3)]
        hit_lens = [10, 20]  # Wrong length
        extractor = FeatureExtractor()

        with pytest.raises(ValueError, match="hit_lens length"):
            extractor.extract_batch(handles, hit_lens=hit_lens)


class TestHistoricalPredictor:
    """Test historical predictor (main implementation)."""

    def test_predict_batch_cold_start(self) -> None:
        """Test prediction with no training data (cold start)."""
        predictor = HistoricalPredictor()
        handles = [KVHandle.create(num_tokens=100) for _ in range(3)]
        features = [PredictorFeatures.from_handle(h) for h in handles]

        predictions = predictor.predict_batch(features)

        assert len(predictions) == 3
        for pred in predictions:
            assert 0 <= pred.reuse_prob <= 1
            assert pred.remaining_lifetime > 0
            assert pred.evict_score > 0
            # Cold start should have lower confidence
            assert pred.confidence < 1.0

    def test_predict_batch_empty(self) -> None:
        """Test error on empty input."""
        predictor = HistoricalPredictor()

        with pytest.raises(ValueError, match="must not be empty"):
            predictor.predict_batch([])

    def test_update_and_predict(self) -> None:
        """Test that predictor learns from updates."""
        predictor = HistoricalPredictor(alpha=0.5)  # Fast learning

        # Create a handle and extract features
        handle = KVHandle.create(num_tokens=200, session_id="session-1")
        features = PredictorFeatures.from_handle(handle)

        # Get initial prediction
        initial_pred = predictor.predict_batch([features])[0]
        initial_lifetime = initial_pred.remaining_lifetime

        # Simulate observing a very long lifetime
        predictor.update(features, actual_lifetime=10000.0)

        # Predict again - should reflect the update
        updated_pred = predictor.predict_batch([features])[0]

        # After observing long lifetime, prediction should increase
        assert updated_pred.remaining_lifetime > initial_lifetime
        # Confidence should increase with more samples
        assert updated_pred.confidence > initial_pred.confidence

    def test_session_aware_prediction(self) -> None:
        """Test that predictor maintains per-session statistics."""
        predictor = HistoricalPredictor()

        # Train on session-1 with long lifetimes
        h1 = KVHandle.create(num_tokens=100, session_id="session-1")
        f1 = PredictorFeatures.from_handle(h1)
        for _ in range(5):
            predictor.update(f1, actual_lifetime=5000.0)

        # Train on session-2 with short lifetimes
        h2 = KVHandle.create(num_tokens=100, session_id="session-2")
        f2 = PredictorFeatures.from_handle(h2)
        for _ in range(5):
            predictor.update(f2, actual_lifetime=100.0)

        # Predict for both sessions
        predictions = predictor.predict_batch([f1, f2])

        # Session-1 should have longer predicted lifetime
        assert predictions[0].remaining_lifetime > predictions[1].remaining_lifetime

    def test_token_size_effect(self) -> None:
        """Test that token count affects predictions."""
        predictor = HistoricalPredictor()

        # Create handles with different token counts
        h_small = KVHandle.create(num_tokens=50)
        h_large = KVHandle.create(num_tokens=500)

        features = [
            PredictorFeatures.from_handle(h_small),
            PredictorFeatures.from_handle(h_large),
        ]

        predictions = predictor.predict_batch(features)

        # Larger KV should generally have longer predicted lifetime
        assert predictions[1].remaining_lifetime > predictions[0].remaining_lifetime

    def test_recency_effect(self) -> None:
        """Test that recent access increases reuse probability."""
        predictor = HistoricalPredictor()

        current_time = time.time()

        # Create two handles: one recently accessed, one not
        h_recent = KVHandle.create(num_tokens=100)
        h_recent.last_access = current_time - 10  # 10 seconds ago

        h_old = KVHandle.create(num_tokens=100)
        h_old.last_access = current_time - 3600  # 1 hour ago

        features = [
            PredictorFeatures.from_handle(h_recent),
            PredictorFeatures.from_handle(h_old),
        ]

        predictions = predictor.predict_batch(features)

        # Recently accessed should have higher reuse probability
        assert predictions[0].reuse_prob > predictions[1].reuse_prob

    def test_get_statistics(self) -> None:
        """Test statistics retrieval."""
        predictor = HistoricalPredictor()

        # Add some training data
        handle = KVHandle.create(num_tokens=100, session_id="session-1")
        features = PredictorFeatures.from_handle(handle)
        predictor.update(features, actual_lifetime=1000.0)

        stats = predictor.get_statistics()

        assert "global" in stats
        assert "session-1" in stats
        assert stats["global"]["sample_count"] > 0
        assert stats["session-1"]["sample_count"] > 0

    def test_get_model_version(self) -> None:
        """Test model version retrieval."""
        predictor = HistoricalPredictor()
        version = predictor.get_model_version()

        assert "historical" in version.lower()
        assert len(version) > 0

    def test_window_size_limit(self) -> None:
        """Test that predictor respects window size limit."""
        predictor = HistoricalPredictor(window_size=3)

        # Create many sessions
        for i in range(10):
            handle = KVHandle.create(num_tokens=100, session_id=f"session-{i}")
            features = PredictorFeatures.from_handle(handle)
            predictor.update(features, actual_lifetime=1000.0)

        stats = predictor.get_statistics()

        # Should not exceed window size + 1 (for global)
        assert len(stats) <= predictor.window_size + 1


class TestPredictorPerformance:
    """Test predictor performance requirements."""

    def test_prediction_latency(self) -> None:
        """Test that prediction meets <1ms latency requirement."""
        predictor = HistoricalPredictor()

        # Create a batch of handles
        handles = [KVHandle.create(num_tokens=100) for _ in range(10)]
        features = [PredictorFeatures.from_handle(h) for h in handles]

        # Measure prediction time
        start = time.time()
        predictions = predictor.predict_batch(features)
        elapsed = time.time() - start

        # Should complete in < 10ms for batch of 10 (generous margin)
        assert elapsed < 0.01
        assert len(predictions) == 10

    def test_update_performance(self) -> None:
        """Test that updates are efficient."""
        predictor = HistoricalPredictor()

        handle = KVHandle.create(num_tokens=100)
        features = PredictorFeatures.from_handle(handle)

        # Measure update time for 100 updates
        start = time.time()
        for _ in range(100):
            predictor.update(features, actual_lifetime=1000.0)
        elapsed = time.time() - start

        # Should complete quickly (< 10ms for 100 updates)
        assert elapsed < 0.01


class TestPredictorIntegration:
    """Integration tests with eviction policies."""

    def test_predictions_usable_by_eviction(self) -> None:
        """Test that predictions can be consumed by eviction policies."""
        predictor = HistoricalPredictor()

        # Create handles with different characteristics
        handles = []
        for i in range(5):
            h = KVHandle.create(num_tokens=100 * (i + 1))
            h.access_count = i
            handles.append(h)

        # Get predictions
        features = [PredictorFeatures.from_handle(h) for h in handles]
        predictions = predictor.predict_batch(features)

        # Sort by evict_score (higher = more evictable)
        sorted_preds = sorted(predictions, key=lambda p: p.evict_score, reverse=True)

        # Verify we can rank handles by evictability
        assert len(sorted_preds) == 5
        assert all(p.evict_score > 0 for p in sorted_preds)

        # First in sorted list should be most evictable
        # (typically low access count)
        most_evictable = sorted_preds[0]
        least_evictable = sorted_preds[-1]

        assert most_evictable.evict_score > least_evictable.evict_score

    def test_predictions_contain_required_fields(self) -> None:
        """Test that predictions contain all Protocol-required fields."""
        predictor = HistoricalPredictor()

        handle = KVHandle.create(num_tokens=100, request_id="req-123")
        features = PredictorFeatures.from_handle(handle)

        predictions = predictor.predict_batch([features])
        pred = predictions[0]

        # Verify all required Protocol fields
        assert pred.handle_id == handle.handle_id
        assert pred.request_id == "req-123"
        assert isinstance(pred.reuse_prob, float)
        assert isinstance(pred.remaining_lifetime, float)
        assert isinstance(pred.evict_score, float)
        assert isinstance(pred.confidence, float)
        assert isinstance(pred.features_summary, dict)
        assert isinstance(pred.predicted_at, float)

        # Verify value ranges
        assert 0 <= pred.reuse_prob <= 1
        assert pred.remaining_lifetime >= 0
        assert 0 <= pred.confidence <= 1



"""Task2.9 Cross-node KV Transfer 完整验收测试套件

覆盖：
1. 压缩/解压链路（bytes payload 的 zlib 压缩，metadata 字段验证）
2. CRC32 校验和（send 端计算，recv 端记录）
3. 拥塞控制/重试（指数退避，retry_count 可观测）
4. 端到端稳定性基准（多次传输，延迟收敛，指标完整）

运行：
    pytest tests/test_kv_transfer_crossnode_advanced.py -v
"""

from __future__ import annotations

import zlib
from typing import Any

import pytest
from sagellm_protocol.kv.enums import CompressionType, TransferStatus

from sagellm_kv_cache import KVHandle, KVTransferEngine, KVTransferFailedError

# ---------------------------------------------------------------------------
# 测试用 backend：支持 bytes/bytearray/list 载荷的内存邮箱
# ---------------------------------------------------------------------------


class _AdvancedMailboxBackend:
    """支持 bytes/bytearray/list 的内存邮箱 backend（用于压缩链路测试）。"""

    def __init__(self, rank: int, mailbox: dict[tuple[int, int], list[Any]]) -> None:
        self._rank = rank
        self._mailbox = mailbox

    def get_rank(self) -> int:
        return self._rank

    def send(self, tensor: Any, dst_rank: int) -> None:
        key = (self._rank, dst_rank)
        if key not in self._mailbox:
            self._mailbox[key] = []
        # 对 bytearray/bytes 做深拷贝，避免引用污染
        if isinstance(tensor, (bytes, bytearray)):
            self._mailbox[key].append(bytes(tensor))
        else:
            self._mailbox[key].append(tensor)

    def recv(self, tensor: Any, src_rank: int) -> None:
        key = (src_rank, self._rank)
        queue = self._mailbox.get(key, [])
        if not queue:
            raise RuntimeError(f"no payload from {src_rank} to {self._rank}")
        payload = queue.pop(0)
        if isinstance(tensor, bytearray):
            tensor[:] = payload
        elif isinstance(tensor, list):
            tensor.clear()
            tensor.extend(payload if isinstance(payload, list) else [payload])
        else:
            raise TypeError(f"Unsupported recv buffer type: {type(tensor)}")


class _TransientFailBackend(_AdvancedMailboxBackend):
    """前 fail_count 次 send 抛出拥塞异常，之后正常。"""

    def __init__(
        self,
        rank: int,
        mailbox: dict[tuple[int, int], list[Any]],
        fail_count: int = 2,
    ) -> None:
        super().__init__(rank, mailbox)
        self._fail_count = fail_count
        self._send_calls = 0

    def send(self, tensor: Any, dst_rank: int) -> None:
        self._send_calls += 1
        if self._send_calls <= self._fail_count:
            raise RuntimeError(f"simulated congestion (attempt {self._send_calls})")
        super().send(tensor, dst_rank)


# ---------------------------------------------------------------------------
# 1. 压缩/解压链路
# ---------------------------------------------------------------------------


class TestCompressionChain:
    """压缩链路：bytes 载荷压缩、元数据字段、解压还原。"""

    def test_compression_metadata_fields(self) -> None:
        """开启压缩后，is_compressed/original_size/compression_type/total_bytes 应正确设置。"""
        mailbox: dict[tuple[int, int], list[Any]] = {}
        engine = KVTransferEngine(
            _AdvancedMailboxBackend(rank=0, mailbox=mailbox),
            enable_compression=True,
            compression_level=6,
        )

        payload = b"x" * 1024  # 1 KiB 可压缩数据
        handle = KVHandle.create(num_tokens=16, request_id="req-compress-meta")

        transfer_id = engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=payload)
        meta = engine.get_transfer_metadata(transfer_id)

        assert meta.is_compressed is True
        assert meta.original_size == 1024
        assert meta.compression_type == CompressionType.ZSTD
        assert meta.total_bytes < meta.original_size, "压缩后应小于原始大小"
        assert meta.status == TransferStatus.COMPLETED

    def test_compressed_payload_is_valid_zlib(self) -> None:
        """实际写入邮箱的载荷应是有效的 zlib 压缩数据。"""
        mailbox: dict[tuple[int, int], list[Any]] = {}
        engine = KVTransferEngine(
            _AdvancedMailboxBackend(rank=0, mailbox=mailbox),
            enable_compression=True,
        )
        payload = b"hello " * 200  # repeating bytes → high compression ratio

        handle = KVHandle.create(num_tokens=8, request_id="req-compress-valid")
        engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=payload)

        # 邮箱中应存放压缩后的数据
        stored = mailbox[(0, 1)][0]
        # 应能被 zlib 解压
        decompressed = zlib.decompress(stored)
        assert decompressed == payload

    def test_compression_ratio_in_metrics(self) -> None:
        """get_transfer_metrics() 中的 compression_ratio 应 > 1 for compressible data."""
        mailbox: dict[tuple[int, int], list[Any]] = {}
        engine = KVTransferEngine(
            _AdvancedMailboxBackend(rank=0, mailbox=mailbox),
            enable_compression=True,
        )
        payload = b"A" * 4096  # highly compressible

        handle = KVHandle.create(num_tokens=32, request_id="req-ratio")
        tid = engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=payload)

        metrics = engine.get_transfer_metrics(tid)
        assert metrics["compression_ratio"] > 1.0
        assert metrics["status"] == "completed"

    def test_decompression_round_trip(self) -> None:
        """send+recv 端对 bytearray 载荷的完整压缩/解压还原。"""
        mailbox: dict[tuple[int, int], list[Any]] = {}
        original = bytearray(b"KV data: " * 300)
        recv_buffer = bytearray(len(original))

        sender = KVTransferEngine(
            _AdvancedMailboxBackend(rank=0, mailbox=mailbox),
            enable_compression=True,
        )
        receiver = KVTransferEngine(
            _AdvancedMailboxBackend(rank=1, mailbox=mailbox),
            enable_compression=True,
        )

        handle = KVHandle.create(num_tokens=16, request_id="req-roundtrip")
        sender.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=original)
        receiver.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=recv_buffer)

        assert bytes(recv_buffer) == bytes(original), "解压后数据应与原始数据一致"

    def test_no_compression_when_disabled(self) -> None:
        """未开启压缩时，is_compressed 应为 False，original_size==0。"""
        mailbox: dict[tuple[int, int], list[Any]] = {}
        engine = KVTransferEngine(
            _AdvancedMailboxBackend(rank=0, mailbox=mailbox),
            enable_compression=False,
        )
        payload = b"plain " * 100
        handle = KVHandle.create(num_tokens=8, request_id="req-no-compress")
        tid = engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=payload)
        meta = engine.get_transfer_metadata(tid)

        assert meta.is_compressed is False
        assert meta.original_size == 0


# ---------------------------------------------------------------------------
# 2. CRC32 校验和验证
# ---------------------------------------------------------------------------


class TestChecksumVerification:
    """CRC32 校验和：send 端设置 checksum，recv 端记录并可在 metrics 中观测。"""

    def test_checksum_set_on_send(self) -> None:
        """send 端开启 enable_checksum 时 metadata.checksum 应为 8 位十六进制串。"""
        mailbox: dict[tuple[int, int], list[Any]] = {}
        engine = KVTransferEngine(
            _AdvancedMailboxBackend(rank=0, mailbox=mailbox),
            enable_checksum=True,
        )
        payload = b"checksum test data"
        handle = KVHandle.create(num_tokens=4, request_id="req-crc")
        tid = engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=payload)
        meta = engine.get_transfer_metadata(tid)

        assert meta.checksum is not None
        assert len(meta.checksum) == 8
        assert all(c in "0123456789abcdef" for c in meta.checksum)

    def test_checksum_matches_crc32(self) -> None:
        """send 端的 checksum 应等于 payload 的 CRC32。"""
        import binascii

        mailbox: dict[tuple[int, int], list[Any]] = {}
        engine = KVTransferEngine(
            _AdvancedMailboxBackend(rank=0, mailbox=mailbox),
            enable_checksum=True,
        )
        payload = b"deterministic payload"
        expected_crc = f"{binascii.crc32(payload) & 0xFFFFFFFF:08x}"

        handle = KVHandle.create(num_tokens=4, request_id="req-crc-check")
        tid = engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=payload)
        meta = engine.get_transfer_metadata(tid)

        assert meta.checksum == expected_crc

    def test_checksum_in_metrics(self) -> None:
        """get_transfer_metrics() 的 'checksum' 字段非空当 enable_checksum=True。"""
        mailbox: dict[tuple[int, int], list[Any]] = {}
        engine = KVTransferEngine(
            _AdvancedMailboxBackend(rank=0, mailbox=mailbox),
            enable_checksum=True,
        )
        payload = b"metrics checksum"
        handle = KVHandle.create(num_tokens=4, request_id="req-crc-metric")
        tid = engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=payload)
        metrics = engine.get_transfer_metrics(tid)

        assert metrics["checksum"] != "", "checksum 字段不应为空"

    def test_no_checksum_when_disabled(self) -> None:
        """未开启 checksum 时 metadata.checksum 应为 None。"""
        mailbox: dict[tuple[int, int], list[Any]] = {}
        engine = KVTransferEngine(
            _AdvancedMailboxBackend(rank=0, mailbox=mailbox),
            enable_checksum=False,
        )
        payload = b"no checksum"
        handle = KVHandle.create(num_tokens=4, request_id="req-no-crc")
        tid = engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=payload)
        meta = engine.get_transfer_metadata(tid)

        assert meta.checksum is None

    def test_compression_and_checksum_combined(self) -> None:
        """压缩 + 校验和同时开启时，meta 字段全部正确填充。"""
        mailbox: dict[tuple[int, int], list[Any]] = {}
        engine = KVTransferEngine(
            _AdvancedMailboxBackend(rank=0, mailbox=mailbox),
            enable_compression=True,
            enable_checksum=True,
        )
        payload = b"combined test " * 100
        handle = KVHandle.create(num_tokens=8, request_id="req-combined")
        tid = engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=payload)
        meta = engine.get_transfer_metadata(tid)

        assert meta.is_compressed is True
        assert meta.original_size == len(payload)
        assert meta.checksum is not None and len(meta.checksum) == 8


# ---------------------------------------------------------------------------
# 3. 拥塞控制：重试与指数退避
# ---------------------------------------------------------------------------


class TestCongestionControlRetry:
    """拥塞控制：transient failure 时重试，retry_count 可观测，超限时抛出异常。"""

    def test_retry_succeeds_after_transient_failures(self) -> None:
        """fail_count=2, max_retries=3 时，第3次 send 成功，transfer 完成。"""
        mailbox: dict[tuple[int, int], list[Any]] = {}
        engine = KVTransferEngine(
            _TransientFailBackend(rank=0, mailbox=mailbox, fail_count=2),
            max_retries=3,
            retry_backoff_ms=1.0,  # 1 ms，避免测试太慢
        )
        handle = KVHandle.create(num_tokens=8, request_id="req-retry-ok")
        payload = [1, 2, 3]

        tid = engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=payload)
        meta = engine.get_transfer_metadata(tid)

        assert meta.status == TransferStatus.COMPLETED
        assert engine.get_retry_count(tid) == 2  # 2 次失败 → 2 次重试

    def test_retry_count_in_metrics(self) -> None:
        """get_transfer_metrics()['retries'] 应等于 get_retry_count()。"""
        mailbox: dict[tuple[int, int], list[Any]] = {}
        engine = KVTransferEngine(
            _TransientFailBackend(rank=0, mailbox=mailbox, fail_count=1),
            max_retries=2,
            retry_backoff_ms=1.0,
        )
        handle = KVHandle.create(num_tokens=4, request_id="req-retry-metric")
        tid = engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=[42])
        metrics = engine.get_transfer_metrics(tid)

        assert metrics["retries"] == engine.get_retry_count(tid)
        assert int(metrics["retries"]) == 1

    def test_no_retry_without_failures(self) -> None:
        """无失败时，retry_count 应为 0。"""
        mailbox: dict[tuple[int, int], list[Any]] = {}
        engine = KVTransferEngine(
            _AdvancedMailboxBackend(rank=0, mailbox=mailbox),
            max_retries=3,
            retry_backoff_ms=1.0,
        )
        handle = KVHandle.create(num_tokens=4, request_id="req-no-retry")
        tid = engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=[99])
        assert engine.get_retry_count(tid) == 0

    def test_raises_when_all_retries_exhausted(self) -> None:
        """fail_count > max_retries 时，应抛出 KVTransferFailedError。"""
        mailbox: dict[tuple[int, int], list[Any]] = {}
        engine = KVTransferEngine(
            _TransientFailBackend(rank=0, mailbox=mailbox, fail_count=5),
            max_retries=2,
            retry_backoff_ms=1.0,
        )
        handle = KVHandle.create(num_tokens=4, request_id="req-retry-fail")

        with pytest.raises(KVTransferFailedError):
            engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=[1])

    def test_zero_max_retries_fails_immediately(self) -> None:
        """max_retries=0 时，任何 send 失败直接传播为 KVTransferFailedError。"""
        mailbox: dict[tuple[int, int], list[Any]] = {}
        engine = KVTransferEngine(
            _TransientFailBackend(rank=0, mailbox=mailbox, fail_count=1),
            max_retries=0,
        )
        handle = KVHandle.create(num_tokens=4, request_id="req-no-retry-fail")

        with pytest.raises(KVTransferFailedError, match="congestion"):
            engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=[1])

    def test_retry_count_cleared_on_clear_completed(self) -> None:
        """clear_completed() 后，retry_count 条目应被清理，不应抛出 KeyError。"""
        mailbox: dict[tuple[int, int], list[Any]] = {}
        engine = KVTransferEngine(
            _AdvancedMailboxBackend(rank=0, mailbox=mailbox),
            max_retries=0,
        )
        handle = KVHandle.create(num_tokens=4, request_id="req-clear")
        tid = engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=[7])
        engine.clear_completed()

        with pytest.raises(KeyError):
            engine.get_retry_count(tid)


# ---------------------------------------------------------------------------
# 4. 端到端稳定性/回归基准
# ---------------------------------------------------------------------------


class TestE2ERegressionBenchmark:
    """端对端稳定性回归：多连续传输延迟收敛，指标字段完整，无内存泄漏。"""

    def test_bulk_transfers_complete(self) -> None:
        """连续 20 次跨节点传输全部完成，completed_count 正确。"""
        mailbox: dict[tuple[int, int], list[Any]] = {}
        engine = KVTransferEngine(_AdvancedMailboxBackend(rank=0, mailbox=mailbox))

        n = 20
        for i in range(n):
            handle = KVHandle.create(num_tokens=8, request_id=f"req-bulk-{i}")
            engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=[i])

        assert engine.get_completed_count() == n

    def test_bulk_transfer_latency_bounded(self) -> None:
        """连续 50 次传输，每次延迟应 < 100 ms（in-memory）。"""
        mailbox: dict[tuple[int, int], list[Any]] = {}
        engine = KVTransferEngine(_AdvancedMailboxBackend(rank=0, mailbox=mailbox))

        for i in range(50):
            handle = KVHandle.create(num_tokens=8, request_id=f"req-latency-{i}")
            tid = engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=bytes(64))
            metrics = engine.get_transfer_metrics(tid)
            assert metrics["latency_ms"] < 100.0, (
                f"Transfer {i} latency too high: {metrics['latency_ms']:.2f} ms"
            )

    def test_metrics_fields_complete(self) -> None:
        """每次传输的 get_transfer_metrics() 应包含所有预期字段。"""
        mailbox: dict[tuple[int, int], list[Any]] = {}
        engine = KVTransferEngine(
            _AdvancedMailboxBackend(rank=0, mailbox=mailbox),
            enable_compression=True,
            enable_checksum=True,
            max_retries=1,
            retry_backoff_ms=1.0,
        )
        handle = KVHandle.create(num_tokens=8, request_id="req-fields")
        tid = engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=b"field test " * 50)
        metrics = engine.get_transfer_metrics(tid)

        required_keys = {
            "bytes",
            "latency_ms",
            "bandwidth_gbps",
            "shards",
            "compression_ratio",
            "retries",
            "checksum",
            "status",
        }
        assert required_keys.issubset(metrics.keys()), (
            f"Missing keys: {required_keys - set(metrics.keys())}"
        )

    def test_status_completed_for_all_transfers(self) -> None:
        """所有成功传输的 status 应为 'completed'。"""
        mailbox: dict[tuple[int, int], list[Any]] = {}
        engine = KVTransferEngine(_AdvancedMailboxBackend(rank=0, mailbox=mailbox))

        tids = []
        for i in range(10):
            handle = KVHandle.create(num_tokens=4, request_id=f"req-status-{i}")
            tid = engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=[i])
            tids.append(tid)

        for tid in tids:
            assert engine.get_transfer_status(tid) == TransferStatus.COMPLETED

    def test_pending_count_zero_after_all_transfers(self) -> None:
        """所有传输完成后，pending_count 应为 0。"""
        mailbox: dict[tuple[int, int], list[Any]] = {}
        engine = KVTransferEngine(_AdvancedMailboxBackend(rank=0, mailbox=mailbox))

        for i in range(5):
            handle = KVHandle.create(num_tokens=4, request_id=f"req-pending-{i}")
            engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=[i])

        assert engine.get_pending_count() == 0

    def test_cross_node_bidirectional_transfer(self) -> None:
        """双向传输：rank0→rank1 + rank1→rank2 各自独立完成。"""
        mailbox: dict[tuple[int, int], list[Any]] = {}
        engine0 = KVTransferEngine(_AdvancedMailboxBackend(rank=0, mailbox=mailbox))
        engine1 = KVTransferEngine(_AdvancedMailboxBackend(rank=1, mailbox=mailbox))

        # rank0 → rank1
        h01 = KVHandle.create(num_tokens=8, request_id="req-bidir-01")
        recv_buf_01: list[Any] = []
        engine0.transfer_kv_handle(h01, src_rank=0, dst_rank=1, tensor=[10, 20])
        engine1.transfer_kv_handle(h01, src_rank=0, dst_rank=1, tensor=recv_buf_01)
        assert recv_buf_01 == [10, 20]

        # rank1 → rank0
        h10 = KVHandle.create(num_tokens=8, request_id="req-bidir-10")
        recv_buf_10: list[Any] = []
        engine1.transfer_kv_handle(h10, src_rank=1, dst_rank=0, tensor=[30, 40])
        engine0.transfer_kv_handle(h10, src_rank=1, dst_rank=0, tensor=recv_buf_10)
        assert recv_buf_10 == [30, 40]

    def test_compression_stability_over_multiple_payloads(self) -> None:
        """多次启用压缩的传输不累计 is_compressed 状态错误，每次元数据独立。"""
        mailbox: dict[tuple[int, int], list[Any]] = {}
        engine = KVTransferEngine(
            _AdvancedMailboxBackend(rank=0, mailbox=mailbox),
            enable_compression=True,
        )
        for i in range(10):
            payload = bytes(range(i % 256)) * 10
            handle = KVHandle.create(num_tokens=8, request_id=f"req-stable-{i}")
            tid = engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=payload)
            meta = engine.get_transfer_metadata(tid)
            assert meta.is_compressed is True
            assert meta.total_bytes > 0
            assert meta.original_size == len(payload)



"""KV Transfer tests for Task2.8/2.9.

Task2.8（baseline）使用真实可执行用例验证 metadata + payload 的最小闭环。
Task2.9（cross-node advanced）仍保留占位测试。
"""

from __future__ import annotations

from typing import Any
from unittest.mock import MagicMock

import pytest
from sagellm_protocol.kv.enums import TransferDirection, TransferStatus

from sagellm_kv_cache import DType, KVHandle, KVTransferEngine, KVTransferFailedError


class _InMemoryBackend:
    """Minimal in-memory backend for cross-node transfer tests."""

    def __init__(self, rank: int, mailbox: dict[tuple[int, int], list[Any]]) -> None:
        self._rank = rank
        self._mailbox = mailbox

    def get_rank(self) -> int:
        return self._rank

    def send(self, tensor: Any, dst_rank: int) -> None:
        key = (self._rank, dst_rank)
        if key not in self._mailbox:
            self._mailbox[key] = []
        self._mailbox[key].append(tensor)

    def recv(self, tensor: list[Any], src_rank: int) -> None:
        key = (src_rank, self._rank)
        queue = self._mailbox.get(key, [])
        if not queue:
            raise RuntimeError(f"no payload from {src_rank} to {self._rank}")
        tensor.clear()
        tensor.extend(queue.pop(0))


class _FailingCongestionBackend(_InMemoryBackend):
    """Backend that simulates congestion on send."""

    def send(self, tensor: Any, dst_rank: int) -> None:
        raise RuntimeError(f"congestion: dst_rank={dst_rank}")


class TestKVTransferProtocol:
    """Task2.8 - KV Transfer Protocol baseline tests."""

    def test_kv_transfer_metadata_schema(self, cpu_config: dict[str, Any]) -> None:
        """Verify metadata schema is populated after transfer."""
        backend = MagicMock()
        backend.get_rank.return_value = 0
        engine = KVTransferEngine(backend)

        handle = KVHandle.create(
            num_tokens=64,
            dtype=DType.FP16,
            device="cpu",
            request_id="req-metadata",
        )
        tensor = MagicMock()

        transfer_id = engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=tensor)
        metadata = engine.get_transfer_metadata(transfer_id)

        assert metadata.transfer_id == transfer_id
        assert metadata.request_id == "req-metadata"
        assert metadata.token_start == 0
        assert metadata.token_end == 64
        assert metadata.total_bytes > 0
        assert metadata.direction == TransferDirection.SEND
        assert metadata.status == TransferStatus.COMPLETED

    def test_send_kv_payload_path(self, cpu_config: dict[str, Any]) -> None:
        """Verify sender path invokes backend send with payload."""
        backend = MagicMock()
        backend.get_rank.return_value = 0
        engine = KVTransferEngine(backend)

        handle = KVHandle.create(num_tokens=32, request_id="req-send")
        tensor = MagicMock()

        transfer_id = engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=tensor)

        backend.send.assert_called_once_with(tensor, 1)
        assert engine.get_transfer_status(transfer_id) == TransferStatus.COMPLETED

    def test_recv_kv_payload_path(self, cpu_config: dict[str, Any]) -> None:
        """Verify receiver path invokes backend recv into payload buffer."""
        backend = MagicMock()
        backend.get_rank.return_value = 1
        engine = KVTransferEngine(backend)

        handle = KVHandle.create(num_tokens=32, request_id="req-recv")
        recv_buffer = MagicMock()

        transfer_id = engine.transfer_kv_handle(
            handle,
            src_rank=0,
            dst_rank=1,
            tensor=recv_buffer,
        )

        backend.recv.assert_called_once_with(recv_buffer, 0)
        assert engine.get_transfer_status(transfer_id) == TransferStatus.COMPLETED

    def test_kv_shard_transfer_baseline(self, cpu_config: dict[str, Any]) -> None:
        """Verify baseline shard-style transfer using two handle slices."""
        backend = MagicMock()
        backend.get_rank.return_value = 0
        engine = KVTransferEngine(backend)

        shard1 = KVHandle.create(num_tokens=16, request_id="req-shard")
        shard1.token_start = 0
        shard1.token_end = 16

        shard2 = KVHandle.create(num_tokens=16, request_id="req-shard")
        shard2.token_start = 16
        shard2.token_end = 32

        tensor1 = MagicMock()
        tensor2 = MagicMock()

        t1 = engine.transfer_kv_handle(shard1, src_rank=0, dst_rank=1, tensor=tensor1)
        t2 = engine.transfer_kv_handle(shard2, src_rank=0, dst_rank=1, tensor=tensor2)

        m1 = engine.get_transfer_metadata(t1)
        m2 = engine.get_transfer_metadata(t2)

        assert (m1.token_start, m1.token_end) == (0, 16)
        assert (m2.token_start, m2.token_end) == (16, 32)
        assert engine.get_completed_count() == 2

    def test_compression_hook_baseline_disabled(self, cpu_config: dict[str, Any]) -> None:
        """Verify baseline keeps compression disabled by default."""
        backend = MagicMock()
        backend.get_rank.return_value = 0
        engine = KVTransferEngine(backend)

        handle = KVHandle.create(num_tokens=8, request_id="req-compress")
        tensor = MagicMock()

        transfer_id = engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=tensor)
        metadata = engine.get_transfer_metadata(transfer_id)

        assert metadata.is_compressed is False
        assert metadata.original_size == 0


class TestCrossNodeKVTransfer:
    """Task2.9 - Cross-node KV Transfer baseline tests."""

    def test_cross_node_transfer_baseline(self, cpu_config: dict[str, Any]) -> None:
        """Verify end-to-end transfer across two in-memory nodes."""
        mailbox: dict[tuple[int, int], list[Any]] = {}
        sender_engine = KVTransferEngine(_InMemoryBackend(rank=0, mailbox=mailbox))
        receiver_engine = KVTransferEngine(_InMemoryBackend(rank=1, mailbox=mailbox))

        handle = KVHandle.create(num_tokens=16, request_id="req-cross-node")
        payload = [1, 2, 3, 4]
        recv_buffer: list[int] = []

        sender_tid = sender_engine.transfer_kv_handle(
            handle, src_rank=0, dst_rank=1, tensor=payload
        )
        receiver_tid = receiver_engine.transfer_kv_handle(
            handle,
            src_rank=0,
            dst_rank=1,
            tensor=recv_buffer,
        )

        assert recv_buffer == payload
        assert sender_engine.get_transfer_status(sender_tid) == TransferStatus.COMPLETED
        assert receiver_engine.get_transfer_status(receiver_tid) == TransferStatus.COMPLETED

    def test_rdma_backend_adapter_contract_baseline(self, cpu_config: dict[str, Any]) -> None:
        """Verify an RDMA-like backend that honors send/recv contract works with engine."""
        mailbox: dict[tuple[int, int], list[Any]] = {}

        class FakeRDMABackend(_InMemoryBackend):
            pass

        sender_engine = KVTransferEngine(FakeRDMABackend(rank=2, mailbox=mailbox))
        receiver_engine = KVTransferEngine(FakeRDMABackend(rank=3, mailbox=mailbox))

        handle = KVHandle.create(num_tokens=8, request_id="req-rdma")
        payload = [7, 8, 9]
        recv_buffer: list[int] = []

        sender_engine.transfer_kv_handle(handle, src_rank=2, dst_rank=3, tensor=payload)
        receiver_engine.transfer_kv_handle(handle, src_rank=2, dst_rank=3, tensor=recv_buffer)

        assert recv_buffer == payload

    def test_compression_integration_baseline_disabled(self, cpu_config: dict[str, Any]) -> None:
        """Verify cross-node baseline keeps compression fields at default values."""
        mailbox: dict[tuple[int, int], list[Any]] = {}
        sender_engine = KVTransferEngine(_InMemoryBackend(rank=4, mailbox=mailbox))

        handle = KVHandle.create(num_tokens=10, request_id="req-compression")
        payload = [42]

        transfer_id = sender_engine.transfer_kv_handle(
            handle, src_rank=4, dst_rank=5, tensor=payload
        )
        metadata = sender_engine.get_transfer_metadata(transfer_id)

        assert metadata.is_compressed is False
        assert metadata.original_size == 0
        assert metadata.checksum is None

    def test_congestion_control_failure_path(self, cpu_config: dict[str, Any]) -> None:
        """Verify send-path congestion surfaces a structured transfer failure."""
        mailbox: dict[tuple[int, int], list[Any]] = {}
        engine = KVTransferEngine(_FailingCongestionBackend(rank=6, mailbox=mailbox))

        handle = KVHandle.create(num_tokens=12, request_id="req-congestion")
        payload = [99]

        with pytest.raises(KVTransferFailedError, match="congestion"):
            engine.transfer_kv_handle(handle, src_rank=6, dst_rank=7, tensor=payload)



"""Tests for KV Transfer Engine MVP implementation.

测试 KV Transfer Engine 与 comm 后端的对接。
"""

from __future__ import annotations

import uuid
from unittest.mock import MagicMock

import pytest
from sagellm_protocol.kv.enums import TransferStatus

from sagellm_kv_cache import KVHandle, KVTransferEngine, KVTransferFailedError


class TestKVTransferEngine:
    """测试 KV Transfer Engine"""

    def test_init_fail_fast_no_backend(self):
        """测试 Fail-Fast：缺少 comm_backend"""
        with pytest.raises(ValueError, match="comm_backend cannot be None"):
            KVTransferEngine(comm_backend=None)

    def test_init_success(self):
        """测试成功初始化"""
        mock_backend = MagicMock()
        engine = KVTransferEngine(comm_backend=mock_backend)

        assert engine._comm_backend is mock_backend
        assert engine.get_pending_count() == 0
        assert engine.get_completed_count() == 0

    def test_init_fail_fast_missing_backend_methods(self):
        """测试 Fail-Fast：backend 缺少必要接口"""

        class IncompleteBackend:
            def get_rank(self):
                return 0

        with pytest.raises(ValueError, match="missing required methods"):
            KVTransferEngine(comm_backend=IncompleteBackend())

    def test_transfer_fail_fast_invalid_src_rank(self):
        """测试 Fail-Fast：无效的 src_rank"""
        mock_backend = MagicMock()
        engine = KVTransferEngine(comm_backend=mock_backend)

        handle = KVHandle(
            handle_id=uuid.uuid4(),
            num_tokens=128,
            token_start=0,
            token_end=128,
        )

        with pytest.raises(ValueError, match="Invalid src_rank"):
            engine.transfer_kv_handle(handle, src_rank=-1, dst_rank=1)

    def test_transfer_fail_fast_invalid_dst_rank(self):
        """测试 Fail-Fast：无效的 dst_rank"""
        mock_backend = MagicMock()
        engine = KVTransferEngine(comm_backend=mock_backend)

        handle = KVHandle(
            handle_id=uuid.uuid4(),
            num_tokens=128,
            token_start=0,
            token_end=128,
        )

        with pytest.raises(ValueError, match="Invalid dst_rank"):
            engine.transfer_kv_handle(handle, src_rank=0, dst_rank=-1)

    def test_transfer_fail_fast_same_rank(self):
        """测试 Fail-Fast：src_rank 和 dst_rank 相同"""
        mock_backend = MagicMock()
        engine = KVTransferEngine(comm_backend=mock_backend)

        handle = KVHandle(
            handle_id=uuid.uuid4(),
            num_tokens=128,
            token_start=0,
            token_end=128,
        )

        with pytest.raises(ValueError, match="cannot be the same"):
            engine.transfer_kv_handle(handle, src_rank=0, dst_rank=0)

    def test_transfer_fail_fast_invalid_current_rank(self):
        """测试 Fail-Fast：当前 rank 既不是源也不是目标"""
        mock_backend = MagicMock()
        mock_backend.get_rank.return_value = 2
        engine = KVTransferEngine(comm_backend=mock_backend)

        handle = KVHandle(
            handle_id=uuid.uuid4(),
            num_tokens=128,
            token_start=0,
            token_end=128,
        )

        with pytest.raises(ValueError, match="neither src.*nor dst"):
            engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1)

    def test_transfer_fail_fast_rank_out_of_world_size(self):
        """测试 Fail-Fast：rank 超出 backend world_size"""
        mock_backend = MagicMock()
        mock_backend.get_rank.return_value = 0
        mock_backend.get_world_size.return_value = 2
        engine = KVTransferEngine(comm_backend=mock_backend)

        handle = KVHandle(
            handle_id=uuid.uuid4(),
            num_tokens=128,
            token_start=0,
            token_end=128,
        )

        with pytest.raises(ValueError, match="Rank out of bounds"):
            engine.transfer_kv_handle(handle, src_rank=0, dst_rank=3, tensor=MagicMock())

    def test_transfer_send_fail_fast_no_tensor(self):
        """测试 Fail-Fast：发送方缺少 tensor"""
        mock_backend = MagicMock()
        mock_backend.get_rank.return_value = 0
        engine = KVTransferEngine(comm_backend=mock_backend)

        handle = KVHandle(
            handle_id=uuid.uuid4(),
            num_tokens=128,
            token_start=0,
            token_end=128,
        )

        with pytest.raises(KVTransferFailedError, match="tensor is required for sender"):
            engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=None)

    def test_transfer_recv_fail_fast_no_tensor(self):
        """测试 Fail-Fast：接收方缺少 tensor buffer"""
        mock_backend = MagicMock()
        mock_backend.get_rank.return_value = 1
        engine = KVTransferEngine(comm_backend=mock_backend)

        handle = KVHandle(
            handle_id=uuid.uuid4(),
            num_tokens=128,
            token_start=0,
            token_end=128,
        )

        with pytest.raises(KVTransferFailedError, match="tensor buffer is required for receiver"):
            engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=None)

    def test_transfer_send_success(self):
        """测试成功发送"""
        mock_backend = MagicMock()
        mock_backend.get_rank.return_value = 0
        engine = KVTransferEngine(comm_backend=mock_backend)

        handle = KVHandle(
            handle_id=uuid.uuid4(),
            num_tokens=128,
            token_start=0,
            token_end=128,
        )

        mock_tensor = MagicMock()
        transfer_id = engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=mock_tensor)

        # 验证发送被调用
        mock_backend.send.assert_called_once_with(mock_tensor, 1)

        # 验证状态
        assert engine.get_pending_count() == 0
        assert engine.get_completed_count() == 1
        assert engine.get_transfer_status(transfer_id) == TransferStatus.COMPLETED

    def test_transfer_recv_success(self):
        """测试成功接收"""
        mock_backend = MagicMock()
        mock_backend.get_rank.return_value = 1
        engine = KVTransferEngine(comm_backend=mock_backend)

        handle = KVHandle(
            handle_id=uuid.uuid4(),
            num_tokens=128,
            token_start=0,
            token_end=128,
        )

        mock_tensor = MagicMock()
        transfer_id = engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=mock_tensor)

        # 验证接收被调用
        mock_backend.recv.assert_called_once_with(mock_tensor, 0)

        # 验证状态
        assert engine.get_pending_count() == 0
        assert engine.get_completed_count() == 1
        assert engine.get_transfer_status(transfer_id) == TransferStatus.COMPLETED

    def test_transfer_send_failure(self):
        """测试发送失败"""
        mock_backend = MagicMock()
        mock_backend.get_rank.return_value = 0
        mock_backend.send.side_effect = RuntimeError("Network error")
        engine = KVTransferEngine(comm_backend=mock_backend)

        handle = KVHandle(
            handle_id=uuid.uuid4(),
            num_tokens=128,
            token_start=0,
            token_end=128,
        )

        mock_tensor = MagicMock()

        with pytest.raises(KVTransferFailedError, match="failed"):
            engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=mock_tensor)

        # 验证失败被记录
        assert engine.get_pending_count() == 0
        assert engine.get_completed_count() == 1

    def test_get_transfer_metadata(self):
        """测试获取传输元数据"""
        mock_backend = MagicMock()
        mock_backend.get_rank.return_value = 0
        engine = KVTransferEngine(comm_backend=mock_backend)

        handle = KVHandle(
            handle_id=uuid.uuid4(),
            num_tokens=128,
            token_start=0,
            token_end=128,
        )

        mock_tensor = MagicMock()
        transfer_id = engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=mock_tensor)

        metadata = engine.get_transfer_metadata(transfer_id)
        assert metadata.transfer_id == transfer_id
        assert metadata.token_start == 0
        assert metadata.token_end == 128

    def test_get_transfer_status_not_found(self):
        """测试获取不存在的传输状态"""
        mock_backend = MagicMock()
        engine = KVTransferEngine(comm_backend=mock_backend)

        with pytest.raises(KeyError, match="not found"):
            engine.get_transfer_status(uuid.uuid4())

    def test_clear_completed(self):
        """测试清理已完成的传输"""
        mock_backend = MagicMock()
        mock_backend.get_rank.return_value = 0
        engine = KVTransferEngine(comm_backend=mock_backend)

        handle = KVHandle(
            handle_id=uuid.uuid4(),
            num_tokens=128,
            token_start=0,
            token_end=128,
        )

        mock_tensor = MagicMock()
        engine.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=mock_tensor)

        assert engine.get_completed_count() == 1

        engine.clear_completed()
        assert engine.get_completed_count() == 0

    def test_get_transfer_metrics(self):
        """测试传输可观测指标输出"""
        mock_backend = MagicMock()
        mock_backend.get_rank.return_value = 0
        engine = KVTransferEngine(comm_backend=mock_backend)

        handle = KVHandle(
            handle_id=uuid.uuid4(),
            num_tokens=64,
            token_start=0,
            token_end=64,
        )

        transfer_id = engine.transfer_kv_handle(
            handle,
            src_rank=0,
            dst_rank=1,
            tensor=MagicMock(),
        )
        metrics = engine.get_transfer_metrics(transfer_id)

        assert metrics["bytes"] > 0
        assert metrics["latency_ms"] >= 0
        assert metrics["bandwidth_gbps"] >= 0
        assert metrics["shards"] == 1
        assert metrics["compression_ratio"] == 1.0
        assert metrics["status"] == TransferStatus.COMPLETED.value



"""KV Pool module for memory management.

This module provides block-level memory allocation and KV pool management.
"""

from __future__ import annotations

from .block_manager import Block, BlockManager
from .kv_pool import KVPool

__all__ = ["Block", "BlockManager", "KVPool"]



"""Task2.9 跨机 KV Transfer：Socket backend 集成测试。"""

from __future__ import annotations

import socket
from contextlib import closing
from typing import Any

from sagellm_kv_cache import KVHandle, KVTransferEngine, SocketTransferBackend


def _free_port() -> int:
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        sock.bind(("127.0.0.1", 0))
        return int(sock.getsockname()[1])


def _build_endpoints(world_size: int) -> dict[int, tuple[str, int]]:
    return {rank: ("127.0.0.1", _free_port()) for rank in range(world_size)}


def test_socket_backend_list_round_trip() -> None:
    endpoints = _build_endpoints(world_size=2)
    backend0 = SocketTransferBackend(rank=0, world_size=2, endpoints=endpoints)
    backend1 = SocketTransferBackend(rank=1, world_size=2, endpoints=endpoints)

    backend0.start()
    backend1.start()
    try:
        sender = KVTransferEngine(backend0)
        receiver = KVTransferEngine(backend1)

        handle = KVHandle.create(num_tokens=16, request_id="req-socket-list")
        recv_buffer: list[Any] = []

        sender.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=[1, 2, 3])
        receiver.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=recv_buffer)

        assert recv_buffer == [1, 2, 3]
    finally:
        backend0.close()
        backend1.close()


def test_socket_backend_bytes_compression_round_trip() -> None:
    endpoints = _build_endpoints(world_size=2)
    backend0 = SocketTransferBackend(rank=0, world_size=2, endpoints=endpoints)
    backend1 = SocketTransferBackend(rank=1, world_size=2, endpoints=endpoints)

    backend0.start()
    backend1.start()
    try:
        sender = KVTransferEngine(backend0, enable_compression=True, enable_checksum=True)
        receiver = KVTransferEngine(backend1, enable_compression=True, enable_checksum=False)

        original = bytearray(b"cross-node-payload-" * 256)
        recv_buffer = bytearray(len(original))
        handle = KVHandle.create(num_tokens=64, request_id="req-socket-bytes")

        sender.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=original)
        receiver.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=recv_buffer)

        assert bytes(recv_buffer) == bytes(original)
    finally:
        backend0.close()
        backend1.close()


def test_socket_backend_metrics_observable() -> None:
    endpoints = _build_endpoints(world_size=2)
    backend0 = SocketTransferBackend(rank=0, world_size=2, endpoints=endpoints)
    backend1 = SocketTransferBackend(rank=1, world_size=2, endpoints=endpoints)

    backend0.start()
    backend1.start()
    try:
        sender = KVTransferEngine(backend0, enable_compression=True, max_retries=1)
        receiver = KVTransferEngine(backend1, enable_compression=True)

        handle = KVHandle.create(num_tokens=8, request_id="req-socket-metrics")
        payload = b"A" * 4096
        recv_buffer = bytearray(len(payload))

        tid = sender.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=payload)
        receiver.transfer_kv_handle(handle, src_rank=0, dst_rank=1, tensor=recv_buffer)

        metrics = sender.get_transfer_metrics(tid)
        assert metrics["bytes"] > 0
        assert metrics["latency_ms"] >= 0.0
        assert metrics["bandwidth_gbps"] >= 0.0
        assert metrics["status"] == "completed"
        assert metrics["compression_ratio"] >= 1.0
    finally:
        backend0.close()
        backend1.close()



"""Integration-oriented tests for KVCacheManager."""

from __future__ import annotations

import pytest

from sagellm_kv_cache.errors import KVBudgetExceededError
from sagellm_kv_cache.kv_cache_manager import KVCacheManager
from sagellm_kv_cache.observability import MetricsCollector


class TestKVCacheManagerPrefix:
    def test_register_and_query_prefix_hit(self) -> None:
        manager = KVCacheManager(max_tokens=1024, block_size=16)
        handle = manager.alloc(num_tokens=64)

        manager.register_prefix((10, 11, 12, 13), handle)
        matched_handle, matched_tokens = manager.query_prefix((10, 11, 12, 13, 14))

        assert matched_handle is not None
        assert matched_handle.handle_id == handle.handle_id
        assert matched_tokens == 4
        assert handle.is_pinned

    def test_query_prefix_reused_blocks(self) -> None:
        manager = KVCacheManager(max_tokens=1024, block_size=16)
        handle = manager.alloc(num_tokens=32)
        manager.register_prefix((1, 2, 3, 4), handle)

        blocks, matched_tokens = manager.get_reused_blocks((1, 2, 3, 4, 5))

        assert matched_tokens == 4
        assert len(blocks) > 0

    def test_prefix_metrics_hit_and_miss(self) -> None:
        collector = MetricsCollector(trace_id="trace-1", request_id="req-1", engine_id="engine-0")
        manager = KVCacheManager(max_tokens=1024, block_size=16, metrics_collector=collector)
        handle = manager.alloc(num_tokens=64)
        manager.register_prefix((1, 2, 3), handle)

        manager.query_prefix((1, 2, 3, 4))
        manager.query_prefix((9, 9, 9))

        metrics = manager.get_prefix_metrics()
        assert metrics["prefix_hits"] == 1
        assert metrics["prefix_lookups"] == 2
        assert metrics["prefix_hit_rate"] == 0.5
        assert metrics["prefix_reuse_tokens"] == 3

        assert collector.get_gauge("prefix_hit_rate") == 0.5
        assert collector.get_gauge("prefix_reuse_tokens") == 3.0

    def test_free_invalidates_prefix_mapping(self) -> None:
        manager = KVCacheManager(max_tokens=1024, block_size=16)
        handle = manager.alloc(num_tokens=64)
        manager.register_prefix((1, 2, 3), handle)

        manager.free(handle)
        matched_handle, matched_tokens = manager.query_prefix((1, 2, 3, 4))

        assert matched_handle is None
        assert matched_tokens == 0


class TestKVCacheManagerEviction:
    def test_alloc_auto_evicts_by_lru(self) -> None:
        manager = KVCacheManager(max_tokens=300)

        h1 = manager.alloc(100, metadata={"request_id": "req-1"})
        h2 = manager.alloc(100, metadata={"request_id": "req-2"})
        h3 = manager.alloc(100, metadata={"request_id": "req-3"})

        h1.last_access = 1.0
        h2.last_access = 2.0
        h3.last_access = 3.0

        h4 = manager.alloc(150, metadata={"request_id": "req-4"})

        remaining_ids = set(manager.kv_pool.handles.keys())
        assert h1.handle_id not in remaining_ids
        assert h2.handle_id not in remaining_ids
        assert h3.handle_id in remaining_ids
        assert h4.handle_id in remaining_ids

    def test_pinned_handle_is_not_evicted(self) -> None:
        manager = KVCacheManager(max_tokens=300)

        h1 = manager.alloc(100, metadata={"request_id": "req-prefix"})
        manager.register_prefix([1, 2, 3], h1)

        h2 = manager.alloc(100, metadata={"request_id": "req-2"})
        h3 = manager.alloc(100, metadata={"request_id": "req-3"})

        h1.last_access = 1.0
        h2.last_access = 2.0
        h3.last_access = 3.0

        manager.alloc(150, metadata={"request_id": "req-4"})

        remaining_ids = set(manager.kv_pool.handles.keys())
        assert h1.handle_id in remaining_ids

    def test_all_pinned_raises_budget_error(self) -> None:
        manager = KVCacheManager(max_tokens=100)

        h1 = manager.alloc(100, metadata={"request_id": "req-1"})
        manager.register_prefix([9, 9, 9], h1)

        with pytest.raises(KVBudgetExceededError):
            manager.alloc(10, metadata={"request_id": "req-2"})

    def test_records_evict_metrics(self) -> None:
        collector = MetricsCollector(
            trace_id="trace-kv-1",
            request_id="req-kv-1",
            engine_id="engine-kv-1",
        )
        manager = KVCacheManager(max_tokens=200, metrics_collector=collector)

        h1 = manager.alloc(100, metadata={"request_id": "req-1"})
        h2 = manager.alloc(100, metadata={"request_id": "req-2"})

        h1.last_access = 1.0
        h2.last_access = 2.0

        manager.alloc(100, metadata={"request_id": "req-3"})

        assert collector.get_counter("evict_count") >= 1
        assert collector.get_gauge("evict_ms") >= 0.0

    def test_respects_single_round_eviction_limit(self) -> None:
        manager = KVCacheManager(max_tokens=300, max_evict_handles_per_round=1)

        h1 = manager.alloc(100)
        h2 = manager.alloc(100)
        h3 = manager.alloc(100)

        h1.last_access = 1.0
        h2.last_access = 2.0
        h3.last_access = 3.0

        with pytest.raises(KVBudgetExceededError):
            manager.alloc(150)



"""真实 Gloo Backend 多进程测试

使用 torch.multiprocessing 进行真实的分布式测试，验证：
1. 真实的 Gloo backend 初始化
2. 多进程环境下的 KV Transfer
3. 跨进程的实际通信

运行方式：
    pytest tests/test_kv_transfer_multiprocess.py -v
"""

from __future__ import annotations

import os
import tempfile
import time
from pathlib import Path
from uuid import uuid4

import pytest
import torch
import torch.multiprocessing as mp
from sagellm_comm import GlooBackend
from sagellm_protocol import DType

from sagellm_kv_cache import KVHandle, KVTransferEngine


def _worker_basic_transfer(rank: int, world_size: int, store_file: str) -> dict:
    """Worker function for basic transfer test

    Args:
        rank: Process rank
        world_size: Total number of processes
        store_file: Path to shared file store

    Returns:
        Test results dictionary
    """
    try:
        # 1. Initialize Gloo backend
        os.environ["RANK"] = str(rank)
        os.environ["WORLD_SIZE"] = str(world_size)
        os.environ["LOCAL_RANK"] = str(rank)
        os.environ["LOCAL_WORLD_SIZE"] = str(world_size)
        os.environ["MASTER_ADDR"] = "127.0.0.1"
        os.environ["MASTER_PORT"] = "29500"

        backend = GlooBackend()
        backend.init(
            rank=rank,
            world_size=world_size,
            master_addr="127.0.0.1",
            master_port=29500,
        )

        # 2. Create KV Transfer Engine
        engine = KVTransferEngine(backend)

        # 3. Create test data
        seq_length = 128
        kv_data = torch.randn(2, 4, seq_length, 8, 64, dtype=torch.float32)
        # Shape: [2, layers, seq, heads, dim] = [2, 4, 128, 8, 64]
        # Size: 2 * 4 * 128 * 8 * 64 * 4 bytes = 1.0 MB

        handle = KVHandle(
            handle_id=uuid4(),
            request_id=f"req_{rank}",
            num_tokens=seq_length,
            dtype=DType.FP32,
            device="cpu",
            memory_ptr=id(kv_data),
            rank=rank,
        )

        # 4. Perform transfer based on rank
        if rank == 0:
            # Rank 0: Send to rank 1
            transfer_id = engine.transfer_kv_handle(
                handle=handle, src_rank=0, dst_rank=1, tensor=kv_data
            )

            status = engine.get_transfer_status(transfer_id)
            metadata = engine.get_transfer_metadata(transfer_id)

            result = {
                "rank": rank,
                "role": "sender",
                "transfer_id": str(transfer_id),
                "status": str(status),
                "source_device": metadata.source_device,
                "target_device": metadata.target_device,
                "total_bytes": metadata.total_bytes,
                "success": True,
            }

        elif rank == 1:
            # Rank 1: Receive from rank 0
            # Wait a bit for sender to prepare
            time.sleep(0.1)

            # Create receive buffer
            recv_buffer = torch.zeros_like(kv_data)

            # Receive
            transfer_id = engine.transfer_kv_handle(
                handle=handle, src_rank=0, dst_rank=1, tensor=recv_buffer
            )

            status = engine.get_transfer_status(transfer_id)
            metadata = engine.get_transfer_metadata(transfer_id)

            result = {
                "rank": rank,
                "role": "receiver",
                "transfer_id": str(transfer_id),
                "status": str(status),
                "source_device": metadata.source_device,
                "target_device": metadata.target_device,
                "total_bytes": metadata.total_bytes,
                "received_data_shape": list(recv_buffer.shape),
                "success": True,
            }

        else:
            result = {"rank": rank, "role": "idle", "success": True}

        # Cleanup
        backend.destroy()

        return result

    except Exception as e:
        return {
            "rank": rank,
            "success": False,
            "error": str(e),
            "error_type": type(e).__name__,
        }


def _worker_batch_transfer(rank: int, world_size: int, store_file: str) -> dict:
    """Worker function for batch transfer test

    Tests multiple transfers from one rank to multiple destinations.
    """
    try:
        # Initialize
        os.environ["RANK"] = str(rank)
        os.environ["WORLD_SIZE"] = str(world_size)
        os.environ["LOCAL_RANK"] = str(rank)
        os.environ["LOCAL_WORLD_SIZE"] = str(world_size)
        os.environ["MASTER_ADDR"] = "127.0.0.1"
        os.environ["MASTER_PORT"] = "29501"  # Different port

        backend = GlooBackend()
        backend.init(
            rank=rank,
            world_size=world_size,
            master_addr="127.0.0.1",
            master_port=29501,
        )
        engine = KVTransferEngine(backend)

        transfer_count = 0
        completed_transfers = []

        if rank == 0:
            # Rank 0: Send to all other ranks
            for dst_rank in range(1, world_size):
                seq_length = 64 + dst_rank * 32
                kv_data = torch.randn(2, 4, seq_length, 8, 64, dtype=torch.float32)

                handle = KVHandle(
                    handle_id=uuid4(),
                    request_id=f"req_0_to_{dst_rank}",
                    num_tokens=seq_length,
                    dtype=DType.FP32,
                    device="cpu",
                    memory_ptr=id(kv_data),
                    rank=0,
                )

                transfer_id = engine.transfer_kv_handle(
                    handle=handle, src_rank=0, dst_rank=dst_rank, tensor=kv_data
                )

                completed_transfers.append({"dst_rank": dst_rank, "transfer_id": str(transfer_id)})
                transfer_count += 1

            result = {
                "rank": rank,
                "role": "sender",
                "transfer_count": transfer_count,
                "transfers": completed_transfers,
                "success": True,
            }

        else:
            # Other ranks: Receive from rank 0
            time.sleep(0.1)

            seq_length = 64 + rank * 32
            recv_buffer = torch.zeros(2, 4, seq_length, 8, 64, dtype=torch.float32)

            handle = KVHandle(
                handle_id=uuid4(),
                request_id=f"req_recv_{rank}",
                num_tokens=seq_length,
                dtype=DType.FP32,
                device="cpu",
                memory_ptr=id(recv_buffer),
                rank=rank,
            )

            transfer_id = engine.transfer_kv_handle(
                handle=handle, src_rank=0, dst_rank=rank, tensor=recv_buffer
            )

            transfer_count = 1
            result = {
                "rank": rank,
                "role": "receiver",
                "transfer_count": transfer_count,
                "transfer_id": str(transfer_id),
                "success": True,
            }

        # Cleanup
        backend.destroy()

        return result

    except Exception as e:
        return {
            "rank": rank,
            "success": False,
            "error": str(e),
            "error_type": type(e).__name__,
        }


# Global wrapper functions for multiprocessing (must be pickle-able)
def _basic_worker_wrapper(rank: int, world_size: int, results_list):
    """Wrapper for basic transfer worker that can be pickled"""
    result = _worker_basic_transfer(rank, world_size, "unused")
    results_list.append(result)


def _batch_worker_wrapper(rank: int, world_size: int, results_list):
    """Wrapper for batch transfer worker that can be pickled"""
    result = _worker_batch_transfer(rank, world_size, "unused")
    results_list.append(result)


def _error_worker_wrapper(rank: int, world_size: int, results_list):
    """Wrapper for error handling worker that can be pickled"""
    result = _worker_with_error(rank, world_size, "unused")
    results_list.append(result)


def _worker_with_error(rank: int, world_size: int, store_file: str) -> dict:
    """Worker function for error handling test"""
    try:
        os.environ["RANK"] = str(rank)
        os.environ["WORLD_SIZE"] = str(world_size)
        os.environ["MASTER_ADDR"] = "127.0.0.1"
        os.environ["MASTER_PORT"] = "29502"

        backend = GlooBackend()
        backend.init(
            rank=rank,
            world_size=world_size,
            master_addr="127.0.0.1",
            master_port=29502,
        )
        engine = KVTransferEngine(backend)

        if rank == 0:
            # Try to send to invalid rank
            kv_data = torch.randn(2, 4, 64, 8, 64)
            handle = KVHandle(
                handle_id=uuid4(),
                request_id="test",
                num_tokens=64,
                dtype=DType.FP32,
                device="cpu",
                memory_ptr=id(kv_data),
                rank=0,
            )

            # This should fail or complete depending on backend behavior
            try:
                engine.transfer_kv_handle(handle=handle, src_rank=0, dst_rank=999, tensor=kv_data)
            except Exception as e:
                # Expected error
                return {
                    "rank": rank,
                    "success": True,
                    "caught_error": str(e),
                    "role": "sender",
                    "transfer_count": 0,
                }

            return {"rank": rank, "success": True, "role": "sender", "transfer_count": 0}
        else:
            # Receiver - just initialize and wait
            return {"rank": rank, "success": True, "role": "receiver"}

    except Exception as e:
        return {"rank": rank, "success": False, "error": str(e)}


@pytest.mark.skipif(not torch.distributed.is_available(), reason="torch.distributed not available")
class TestKVTransferMultiprocess:
    """Multi-process tests with real Gloo backend"""

    def test_basic_transfer_2_processes(self, tmp_path: Path):
        """Test basic KV transfer with 2 processes"""
        world_size = 2

        # Spawn processes using multiprocessing
        mp.set_start_method("spawn", force=True)
        manager = mp.Manager()
        results = manager.list()

        processes = []
        for rank in range(world_size):
            p = mp.Process(target=_basic_worker_wrapper, args=(rank, world_size, results))
            p.start()
            processes.append(p)

        # Wait for completion
        for p in processes:
            p.join(timeout=30)
            if p.exitcode != 0:
                raise RuntimeError(f"Process failed with exit code {p.exitcode}")

        # Check results
        assert len(results) == world_size, f"Expected {world_size} results, got {len(results)}"

        for result in results:
            assert result["success"], f"Rank {result['rank']} failed: {result.get('error')}"

        # Verify sender and receiver
        sender = next(r for r in results if r.get("role") == "sender")
        receiver = next(r for r in results if r.get("role") == "receiver")

        assert sender["rank"] == 0
        assert receiver["rank"] == 1
        assert "TransferStatus.COMPLETED" in sender["status"]
        assert "TransferStatus.COMPLETED" in receiver["status"]

        print("\n✅ Basic 2-process transfer test passed!")
        print(f"  - Sender (rank 0): {sender['total_bytes']} bytes transferred")
        print(f"  - Receiver (rank 1): Received data shape {receiver['received_data_shape']}")

    def test_batch_transfer_4_processes(self, tmp_path: Path):
        """Test batch KV transfer with 4 processes"""
        world_size = 4

        # Spawn processes
        mp.set_start_method("spawn", force=True)
        manager = mp.Manager()
        results = manager.list()

        processes = []
        for rank in range(world_size):
            p = mp.Process(target=_batch_worker_wrapper, args=(rank, world_size, results))
            p.start()
            processes.append(p)

        # Wait for completion
        for p in processes:
            p.join(timeout=60)
            if p.exitcode != 0:
                raise RuntimeError(f"Process failed with exit code {p.exitcode}")

        # Check results
        assert len(results) == world_size, f"Expected {world_size} results, got {len(results)}"

        for result in results:
            assert result["success"], f"Rank {result['rank']} failed: {result.get('error')}"

        # Verify sender sent to all receivers
        sender = next(r for r in results if r.get("role") == "sender")
        receivers = [r for r in results if r.get("role") == "receiver"]

        assert sender["rank"] == 0
        assert sender["transfer_count"] == world_size - 1  # Sent to all others
        assert len(receivers) == world_size - 1

        print(f"\n✅ Batch {world_size}-process transfer test passed!")
        print(f"  - Sender (rank 0): {sender['transfer_count']} transfers")
        for r in receivers:
            print(f"  - Receiver (rank {r['rank']}): Received {r['transfer_count']} transfer")

    def test_error_handling_multiprocess(self, tmp_path: Path):
        """Test error handling in multi-process KV transfer"""
        world_size = 2

        # Spawn processes
        mp.set_start_method("spawn", force=True)
        manager = mp.Manager()
        results = manager.list()

        processes = []
        for rank in range(world_size):
            p = mp.Process(target=_error_worker_wrapper, args=(rank, world_size, results))
            p.start()
            processes.append(p)

        # Wait for completion
        for p in processes:
            p.join(timeout=60)
            if p.exitcode != 0:
                raise RuntimeError(f"Process failed with exit code {p.exitcode}")

        # Check results
        assert len(results) == world_size

        for result in results:
            assert result["success"], f"Rank {result['rank']} failed unexpectedly"

        sender = next(r for r in results if r.get("role") == "sender")
        next(r for r in results if r.get("role") == "receiver")

        # Sender should handle error gracefully
        assert sender["rank"] == 0
        assert sender["transfer_count"] == 0, "Should not have successful transfers"

        print("\n✅ Error handling multiprocess test passed!")
        print("  - Sender handled error gracefully")
        print("  - Receiver initialized successfully")


@pytest.mark.skipif(not torch.distributed.is_available(), reason="torch.distributed not available")
def test_real_gloo_backend_initialization():
    """Test that real Gloo backend can be initialized in single process"""
    # This is a simpler test that doesn't require multiprocessing
    # Useful for debugging

    with tempfile.TemporaryDirectory():
        # Create backend and initialize (it will call dist.init_process_group internally)
        backend = GlooBackend()
        backend.init(
            rank=0,
            world_size=1,
            master_addr="127.0.0.1",
            master_port=29503,
        )

        # Create engine
        KVTransferEngine(backend)

        # Verify basic properties
        assert backend.get_rank() == 0
        assert backend.get_world_size() == 1

        # Cleanup
        backend.destroy()

        print("\n✅ Real Gloo backend initialization test passed!")


if __name__ == "__main__":
    # Can run directly for debugging
    import sys

    print("Running multiprocess KV Transfer tests...")
    print("Note: For full test suite, use: pytest tests/test_kv_transfer_multiprocess.py -v")

    # Run simple initialization test
    if torch.distributed.is_available():
        test_real_gloo_backend_initialization()
    else:
        print("❌ torch.distributed not available, skipping tests")
        sys.exit(1)



"""Tests for Scheduler Bridge (Task 2.6/2.7)"""

from sagellm_kv_cache.load_aware_scheduler import PlacementStrategy
from sagellm_kv_cache.models import EvictionPolicy, RequestType, SchedulerRequest
from sagellm_kv_cache.scheduler_bridge import SchedulerBridge


class TestSchedulerBridge:
    """Test suite for SchedulerBridge."""

    def test_initialization(self):
        """Test basic initialization."""
        bridge = SchedulerBridge(max_tokens=1024, block_size=16)

        assert bridge.max_tokens == 1024
        assert bridge.block_size == 16
        assert bridge.kv_pool is not None
        assert bridge.prefix_cache is not None

    def test_allocate_request(self):
        """Test allocating KV cache for a request."""
        bridge = SchedulerBridge(max_tokens=1024)

        handle = bridge.allocate_request("req1", num_tokens=100)

        assert handle.num_tokens == 100
        assert "req1" in bridge._request_handles

        # Check pool stats
        stats = bridge.get_stats()
        assert stats["pool"]["allocated_tokens"] == 100
        assert stats["num_requests"] == 1

    def test_free_request(self):
        """Test freeing KV cache for a request."""
        bridge = SchedulerBridge(max_tokens=1024)

        bridge.allocate_request("req1", num_tokens=100)
        bridge.free_request("req1")

        assert "req1" not in bridge._request_handles
        stats = bridge.get_stats()
        assert stats["pool"]["allocated_tokens"] == 0

    def test_plan_single_request(self):
        """Test creating plan for single request."""
        bridge = SchedulerBridge(max_tokens=1024)

        req = SchedulerRequest(
            request_id="req1",
            trace_id="trace1",
            request_type=RequestType.PREFILL,
            prompt_len=100,
            max_new_tokens=50,
            kv_tokens_required=150,
        )

        plan = bridge.plan([req])

        assert plan.batch_size == 1
        assert plan.total_kv_tokens >= 150
        assert "req1" in plan.placement

    def test_plan_multiple_requests(self):
        """Test creating plan for multiple requests."""
        bridge = SchedulerBridge(max_tokens=2048)

        reqs = [
            SchedulerRequest(
                request_id=f"req{i}",
                trace_id=f"trace{i}",
                request_type=RequestType.PREFILL,
                prompt_len=50,
                max_new_tokens=50,
                kv_tokens_required=100,
            )
            for i in range(5)
        ]

        plan = bridge.plan(reqs)

        assert plan.batch_size == 5
        assert plan.total_kv_tokens >= 500

    def test_plan_exceeds_budget(self):
        """Test plan when budget is exceeded."""
        bridge = SchedulerBridge(max_tokens=200)

        # First request uses most of budget
        req1 = SchedulerRequest(
            request_id="req1",
            trace_id="trace1",
            request_type=RequestType.PREFILL,
            prompt_len=150,
            max_new_tokens=50,
            kv_tokens_required=180,
        )

        # Second request won't fit
        req2 = SchedulerRequest(
            request_id="req2",
            trace_id="trace2",
            request_type=RequestType.PREFILL,
            prompt_len=50,
            max_new_tokens=50,
            kv_tokens_required=100,
        )

        bridge.plan([req1, req2])

        # At least one should fail to allocate due to budget
        assert len(bridge._request_handles) < 2

    def test_eviction_triggered(self):
        """Test that eviction is triggered when needed."""
        bridge = SchedulerBridge(max_tokens=300, eviction_policy=EvictionPolicy.LRU)

        # Allocate three requests
        bridge.allocate_request("req1", 100)
        h2 = bridge.allocate_request("req2", 100)
        h3 = bridge.allocate_request("req3", 100)

        # Touch h2 and h3 to make them more recent
        h2.touch()
        h3.touch()

        # Now try to allocate req4 (needs eviction)
        bridge.allocate_request("req4", 150)

        # req1 should be evicted (oldest)
        assert "req1" not in bridge._request_handles
        assert "req4" in bridge._request_handles

    def test_eviction_respects_pinned(self):
        """Test that pinned handles are not evicted."""
        bridge = SchedulerBridge(max_tokens=300)

        h1 = bridge.allocate_request("req1", 100)
        bridge.allocate_request("req2", 100)
        bridge.allocate_request("req3", 100)

        # Pin h1
        bridge.kv_pool.pin(h1)

        # Try to allocate req4 (needs eviction)
        bridge.allocate_request("req4", 150)

        # h1 should NOT be evicted (pinned)
        assert "req1" in bridge._request_handles
        # h2 or h3 should be evicted
        assert "req4" in bridge._request_handles

    def test_reset(self):
        """Test resetting bridge state."""
        bridge = SchedulerBridge(max_tokens=1024)

        bridge.allocate_request("req1", 100)
        bridge.allocate_request("req2", 100)

        assert len(bridge._request_handles) == 2

        bridge.reset()

        assert len(bridge._request_handles) == 0
        stats = bridge.get_stats()
        assert stats["pool"]["allocated_tokens"] == 0

    def test_prefix_cache_disabled(self):
        """Test bridge with prefix cache disabled."""
        bridge = SchedulerBridge(max_tokens=1024, enable_prefix_cache=False)

        assert bridge.prefix_cache is None
        stats = bridge.get_stats()
        assert "prefix_cache" not in stats

    def test_device_affinity(self):
        """Test device affinity in plan."""
        bridge = SchedulerBridge(max_tokens=1024)

        req = SchedulerRequest(
            request_id="req1",
            trace_id="trace1",
            request_type=RequestType.PREFILL,
            prompt_len=100,
            max_new_tokens=50,
            kv_tokens_required=150,
            device_affinity="cuda:1",
        )

        plan = bridge.plan([req])

        assert plan.placement["req1"] == "cuda:1"

    def test_repr(self):
        """Test string representation."""
        bridge = SchedulerBridge(max_tokens=1024, block_size=16)
        repr_str = repr(bridge)

        assert "SchedulerBridge" in repr_str
        assert "max_tokens=1024" in repr_str
        assert "block_size=16" in repr_str

    def test_load_aware_placement_lowest_kv(self):
        """Test load-aware placement picks lower KV occupancy device."""
        bridge = SchedulerBridge(
            max_tokens=2000,
            enable_load_awareness=True,
            placement_strategy=PlacementStrategy.LOWEST_KV,
            devices=["cuda:0", "cuda:1"],
        )

        bridge.allocate_request("seed0", num_tokens=900, device="cuda:0")
        bridge.allocate_request("seed1", num_tokens=100, device="cuda:1")

        req = SchedulerRequest(
            request_id="req-lowest-kv",
            trace_id="trace-lowest-kv",
            request_type=RequestType.PREFILL,
            prompt_len=32,
            max_new_tokens=16,
            kv_tokens_required=48,
        )
        plan = bridge.plan([req])

        assert plan.placement["req-lowest-kv"] == "cuda:1"

    def test_load_metrics_exposed_in_stats(self):
        """Test per-device metrics are visible in bridge stats."""
        bridge = SchedulerBridge(
            max_tokens=1024,
            enable_load_awareness=True,
            devices=["cuda:0", "cuda:1"],
        )
        bridge.allocate_request("req-metrics", num_tokens=128, device="cuda:1")

        stats = bridge.get_stats()

        assert "load_metrics" in stats
        assert "placement" in stats
        assert "cuda:1" in stats["load_metrics"]
        assert stats["load_metrics"]["cuda:1"]["active_requests"] >= 1



"""Tests for KV Cache access statistics collector"""

from __future__ import annotations

import json
import tempfile
import time
from pathlib import Path

from sagellm_kv_cache.profiling import AccessStatsCollector


class TestAccessStatsCollector:
    """测试 AccessStatsCollector 类"""

    def test_init(self) -> None:
        """测试初始化"""
        collector = AccessStatsCollector()
        assert collector.total_accesses == 0
        assert collector.cache_hits == 0
        assert collector.cache_misses == 0
        assert len(collector.access_times) == 0
        assert len(collector.access_count) == 0

    def test_record_access_hit(self) -> None:
        """测试记录命中访问"""
        collector = AccessStatsCollector()
        collector.record_access("block_1", is_hit=True)

        assert collector.total_accesses == 1
        assert collector.cache_hits == 1
        assert collector.cache_misses == 0
        assert collector.access_count["block_1"] == 1
        assert len(collector.access_times["block_1"]) == 1

    def test_record_access_miss(self) -> None:
        """测试记录未命中访问"""
        collector = AccessStatsCollector()
        collector.record_access("block_1", is_hit=False)

        assert collector.total_accesses == 1
        assert collector.cache_hits == 0
        assert collector.cache_misses == 1

    def test_multiple_accesses(self) -> None:
        """测试多次访问"""
        collector = AccessStatsCollector()

        for i in range(10):
            collector.record_access(f"block_{i % 3}", is_hit=True)
            time.sleep(0.001)  # 确保时间间隔

        assert collector.total_accesses == 10
        assert len(collector.access_count) == 3
        assert collector.access_count["block_0"] == 4
        assert collector.access_count["block_1"] == 3
        assert collector.access_count["block_2"] == 3

    def test_get_access_intervals(self) -> None:
        """测试获取访问时间间隔"""
        collector = AccessStatsCollector()

        # 单次访问，无间隔
        collector.record_access("block_1", is_hit=True)
        intervals = collector.get_access_intervals("block_1")
        assert intervals == []

        # 多次访问
        time.sleep(0.01)
        collector.record_access("block_1", is_hit=True)
        time.sleep(0.01)
        collector.record_access("block_1", is_hit=True)

        intervals = collector.get_access_intervals("block_1")
        assert len(intervals) == 2
        assert all(interval > 0 for interval in intervals)

    def test_get_all_intervals(self) -> None:
        """测试获取所有访问间隔"""
        collector = AccessStatsCollector()

        for _ in range(5):
            collector.record_access("block_1", is_hit=True)
            time.sleep(0.001)

        for _ in range(3):
            collector.record_access("block_2", is_hit=True)
            time.sleep(0.001)

        all_intervals = collector.get_all_intervals()
        assert len(all_intervals) == 6  # 4 + 2

    def test_get_hit_rate(self) -> None:
        """测试命中率计算"""
        collector = AccessStatsCollector()

        # 无访问时命中率为 0
        assert collector.get_hit_rate() == 0.0

        # 记录一些访问
        for _ in range(8):
            collector.record_access("block_1", is_hit=True)

        for _ in range(2):
            collector.record_access("block_2", is_hit=False)

        assert collector.get_hit_rate() == 0.8

    def test_get_top_blocks(self) -> None:
        """测试获取访问频率最高的 block"""
        collector = AccessStatsCollector()

        # 生成不同频率的访问
        for i in range(1, 6):
            for _ in range(i):
                collector.record_access(f"block_{i}", is_hit=True)

        top_3 = collector.get_top_blocks(3)
        assert len(top_3) == 3
        assert top_3[0][0] == "block_5"
        assert top_3[0][1] == 5
        assert top_3[1][0] == "block_4"
        assert top_3[1][1] == 4

    def test_get_stats_summary(self) -> None:
        """测试统计摘要"""
        collector = AccessStatsCollector()

        for i in range(10):
            collector.record_access(f"block_{i % 3}", is_hit=i % 2 == 0)
            time.sleep(0.001)

        summary = collector.get_stats_summary()

        assert summary["total_accesses"] == 10
        assert summary["unique_blocks"] == 3
        assert "hit_rate" in summary
        assert "cache_hits" in summary
        assert "cache_misses" in summary
        assert "average_access_interval" in summary
        assert "duration_seconds" in summary
        assert "top_10_blocks" in summary

    def test_export_stats(self) -> None:
        """测试导出统计数据"""
        collector = AccessStatsCollector()

        for i in range(5):
            collector.record_access(f"block_{i % 2}", is_hit=True)
            time.sleep(0.001)

        with tempfile.TemporaryDirectory() as tmpdir:
            output_path = Path(tmpdir) / "test_stats.json"
            collector.export_stats(output_path)

            assert output_path.exists()

            # 验证 JSON 内容
            with open(output_path, encoding="utf-8") as f:
                data = json.load(f)

            assert data["total_accesses"] == 5
            assert data["unique_blocks"] == 2
            assert "access_intervals" in data
            assert "block_details" in data

    def test_reset(self) -> None:
        """测试重置功能"""
        collector = AccessStatsCollector()

        for i in range(10):
            collector.record_access(f"block_{i}", is_hit=True)

        assert collector.total_accesses == 10

        collector.reset()

        assert collector.total_accesses == 0
        assert collector.cache_hits == 0
        assert collector.cache_misses == 0
        assert len(collector.access_times) == 0
        assert len(collector.access_count) == 0

    def test_block_details_export(self) -> None:
        """测试 block 详细信息的导出"""
        collector = AccessStatsCollector()

        # 模拟多次访问
        collector.record_access("block_1", is_hit=True)
        time.sleep(0.01)
        collector.record_access("block_1", is_hit=True)
        time.sleep(0.01)
        collector.record_access("block_2", is_hit=False)

        with tempfile.TemporaryDirectory() as tmpdir:
            output_path = Path(tmpdir) / "test_details.json"
            collector.export_stats(output_path)

            with open(output_path, encoding="utf-8") as f:
                data = json.load(f)

            assert "block_1" in data["block_details"]
            assert "block_2" in data["block_details"]

            block_1_details = data["block_details"]["block_1"]
            assert block_1_details["access_count"] == 2
            assert "first_access" in block_1_details
            assert "last_access" in block_1_details
            assert "intervals" in block_1_details
            assert len(block_1_details["intervals"]) == 1
